/* floatator.c */

#include "problem/model.h"

#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <unistd.h>

struct MasterPackage *ms_pkg;

int main( int argc, char **argv, char **env )
{
    int connection;
    struct MasterPackage pkg, new_pkg;
    
    init_MasterPackage( &pkg, argc, argv, env );
    while( pkg.running )
    {
	ms_pkg = &pkg;
	connection        = create_socket      ( pkg.socket_name );
	pkg.pid_interface = spawn_interface    ( pkg.interface_name, argv, env );
	pkg.s_interface   = accept_from_socket ( connection );
	new_pkg           = floatator_main_loop( &pkg );
	purge_socket( &pkg );
	pkg = new_pkg;
    }
    return 0;
}

void init_MasterPackage
( 
    struct MasterPackage *pkg, 
    int argc, char **argv, char **env 
)
{
    pkg->socket_name = strdup( SOCKET_NAME );
    pkg->interface_name = strdup( INTERFACE_NAME );
    pkg->t = 0.0;
    pkg->running = TRUE;
}

int create_socket( char *socket_name )
{
    int s, success;
    struct sockaddr_un adr;

    unlink( socket_name );

    s = socket( AF_UNIX, SOCK_STREAM, 0 );
    if( s == -1 )
    {
	fprintf( stderr, "A socket could not be created.\n" );
	exit( 1 );
    }
    
    strcpy( adr.sun_path, socket_name );
    adr.sun_family = AF_UNIX;

    success = bind
    ( 
        s, (struct sockaddr *)&adr, 
	strlen(adr.sun_path)+sizeof(adr.sun_family)
    );
    if( success == -1 )
    {
	fprintf
	( 
	    stderr, 
	    "The socket could not be bound to the name \"%s\".\n",  
	    socket_name
	);
	exit( 1 );
    }

    success = listen( s, 1 );
    if( success == -1 )
    {
	fprintf( stderr, "The socket could not be listen to.\n" );
	exit( 1 );
    }

    return s;
}

int spawn_interface( char *interface_name, char **argv, char **env )
{
    char *save_argv_0;
    int pid;
    
    save_argv_0 = argv[0];
    argv[0] = interface_name;
    
    pid = fork();
    
    if( pid == 0 )
    {
#if 1
	/* This is the child process */
        execve( interface_name, argv, env );
#else
	exit( system ("ddd floatator_X_interface"));
#endif
    } else if( pid == (pid_t) - 1 ) {
	fprintf( stderr, "A process could not be spawned.\n" );
	exit( 1 );
    }
    
    argv[0] = save_argv_0;
    
    return pid;
}

int accept_from_socket( int connection )
{
    int s_interface;
    
    s_interface = accept( connection, NULL, NULL );
    if( s_interface == -1 )
    {
	fprintf( stderr, "The connection from the socket could not be accepted.\n" );
	exit( 1 );
    }
    
    close( connection );
    
    return s_interface;
}

void purge_socket( struct MasterPackage *pkg )
{
    shutdown( pkg->s_interface, 2 ); /* no data is to be sent or received */
    unlink( pkg->socket_name );
}
