/* io.c */

#include "problem/init.h"
#include "problem/model.h"

#include <math.h>

void **AllocMatrix( int imin, int imax, int jmin, int jmax, int size )
{
    int i;
    void **result;
    
    result = (void **)malloc( (imax-imin+1) * sizeof(void*) );
    if( result == 0 )
    {
	fprintf( stderr, "Out of memory" );
	exit( 1 );
    }    
    result -= imin;
    
    for( i = imin; i <= imax; i++ )
    {
	result[i] = malloc( (jmax-jmin+1) * size );
	memset( result[i], 0, (jmax-jmin+1) * size );
	if( result[i] == 0 )
	{
	    fprintf( stderr, "Out of memory" );
	    exit( 1 );
	}    
	result[i] = (char *)result[i] - jmin*size;
    }
    
    return result;
}

void FreeMatrix
( 
    void **m, 
    int imin, int imax, 
    int jmin, int jmax, 
    int size 
)
{
    int i;
    for( i = imin; i <= imax; i++ )
	free( (char *)m[i] + jmin*size );
    free( m + imin );    
}

void alloc_if_necessary
( 
    struct DataPackage *pkg, 
    int alloc_psiZeta, 
    int alloc_fGRhs
)
{
    if( pkg->FLAG == 0 )
	pkg->FLAG = (int **)AllocMatrix
	( 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(int) 
	);
    if( pkg->U == 0 )
	pkg->U = (REAL **)AllocMatrix
	( 
	    pkg->imin-2, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    if( pkg->V == 0 )
	pkg->V = (REAL **)AllocMatrix
	( 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-2, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    if( pkg->P == 0 )
	pkg->P = (REAL **)AllocMatrix
	( 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
	if( pkg->PSI == 0 )
	    pkg->PSI = (REAL **)AllocMatrix
	    ( 
		pkg->imin-1, pkg->imax+1, 
		pkg->jmin-1, pkg->jmax+1, 
		sizeof(REAL) 
	    );
	if( pkg->ZETA == 0 )
	    pkg->ZETA = (REAL **)AllocMatrix
	    ( 
		pkg->imin-1, pkg->imax+1, 
		pkg->jmin-1, pkg->jmax+1, 
		sizeof(REAL) 
	    );
	if( pkg->F == 0 )
	    pkg->F = (REAL **)AllocMatrix
	    ( 
		pkg->imin-2, pkg->imax+1, 
		pkg->jmin-1, pkg->jmax+1, 
		sizeof(REAL) 
	    );
	if( pkg->G == 0 )
	    pkg->G = (REAL **)AllocMatrix
	    ( 
		pkg->imin-1, pkg->imax+1, 
		pkg->jmin-2, pkg->jmax+1, 
		sizeof(REAL) 
	    );
	if( pkg->RHS == 0 )
	    pkg->RHS = (REAL **)AllocMatrix
	    ( 
		pkg->imin-1, pkg->imax+1, 
		pkg->jmin-1, pkg->jmax+1, 
		sizeof(REAL) 
	    );
}

void free_if_necessary( struct DataPackage *pkg )
{
    if( pkg->FLAG != 0 )
	FreeMatrix
	(
	    (void **)pkg->FLAG,  
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(int) 
	);
    pkg->FLAG = 0;
    if( pkg->U != 0 )
	FreeMatrix
	( 
	    (void **)pkg->U, 
	    pkg->imin-2, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->U = 0;
    if( pkg->V != 0 )
	FreeMatrix
	( 
	    (void **)pkg->V, 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-2, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->V = 0;
    if( pkg->P != 0 )
	FreeMatrix
	( 
	    (void **)pkg->P, 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->P = 0;
    if( pkg->PSI != 0 )
	FreeMatrix
	( 
	    (void **)pkg->PSI, 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->PSI = 0;
    if( pkg->ZETA != 0 )
	FreeMatrix
	( 
	    (void **)pkg->ZETA, 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->ZETA = 0;
    if( pkg->F != 0 )
	FreeMatrix
	( 
	    (void **)pkg->F, 
	    pkg->imin-2, pkg->imax+1, 
	    pkg->jmin-1, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->F = 0;
    if( pkg->G != 0 )
	FreeMatrix
	( 
	    (void **)pkg->G, 
	    pkg->imin-1, pkg->imax+1, 
	    pkg->jmin-2, pkg->jmax+1, 
	    sizeof(REAL) 
	);
    pkg->G = 0;
    if( pkg->RHS != 0 )
	FreeMatrix
	( 
	    (void **)pkg->RHS, 
	    pkg->imin, pkg->imax, 
	    pkg->jmin, pkg->jmax, 
	    sizeof(REAL) 
	);
    pkg->RHS = 0;	
}

static void free_linked_list ( struct Particle *p )
{
    while( p )
	{
	    struct Particle *next_p = p->next;
	    free( p );
	    p = next_p;
	}
}

static void free_Particles( struct DataPackage *pkg )
{
    int n;
    for( n = 0; n < FLUID_FLAVORS; n++ )
    {
	free_linked_list( pkg->fluids[n].first );
	pkg->fluids[n].first = 0;

	pkg->fluids[n].injection = 0;
    }
    for( n = 0; n < PARTICLE_FLAVORS; n++ )
    {
	free_linked_list( pkg->particles[n].first ); 
	pkg->particles[n].first = 0;

	free_linked_list( pkg->particles[n].injection ); 
	pkg->particles[n].injection = 0;
    }
}

void read_DataPackage_from_file
( 
    struct DataPackage *pkg,
    FILE *f, 
    int read_psiZeta 
)
{
    struct FileHeader header;
    struct ParticleInFile pf;
    struct Particle *p;
    int i;
    int imin = pkg->imin;
    int jmin = pkg->jmin;
    int imax = pkg->imax;
    int jmax = pkg->jmax;
    
    fseek( f, FILE_HEADER_POS, SEEK_SET );
    fread( &header, sizeof(header), 1, f );
    model.parms = header.parms;
    imin = jmin = pkg->imin = pkg->jmin = 1;
    imax = pkg->imax = header.imax;
    jmax = pkg->jmax = header.jmax;
    pkg->xmin = pkg->ymin = 0.0;
    pkg->xmax = pkg->parms.xlength;
    pkg->ymax = pkg->parms.ylength;
    pkg->time = header.t;
    pkg->delt = header.delt;
    pkg->res  = header.res;
    pkg->it   = header.it;
    
    free_Particles( pkg );
    alloc_if_necessary( pkg, read_psiZeta, FALSE );
    
    for( i = imin; i <= imax; i++ )
    {
	fseek( f, U_POS(i, jmin, header), SEEK_SET );
 	fread( pkg->U[i]+jmin, sizeof(REAL), jmax-jmin+1, f );
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, V_POS(i, jmin, header), SEEK_SET );
 	fread( pkg->V[i]+jmin, sizeof(REAL), jmax-jmin+1, f );
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, P_POS(i, jmin, header), SEEK_SET );
 	fread( pkg->P[i]+jmin, sizeof(REAL), jmax-jmin+1, f );
    }
    if( read_psiZeta )
    {
	for( i = pkg->imin; i <= pkg->imax; i++ )
	{
	    fseek( f, PSI_POS(i, pkg->jmin, header), SEEK_SET );
	    fread( pkg->PSI[i]+jmin, sizeof(REAL), jmax-jmin+1, f );
	}
	for( i = pkg->imin; i <= pkg->imax; i++ )
	{
	    fseek( f, ZETA_POS(i, jmin, header), SEEK_SET );
	    fread( pkg->ZETA[i]+jmin, sizeof(REAL), jmax-jmin+1, f );
	}
    }
    for( i = imin; i <= imax; i++ )
    {
	fseek( f, FLAG_POS(i, jmin, header), SEEK_SET );
 	fread( pkg->FLAG[i]+jmin, sizeof(int), jmax-jmin+1, f );
    }
    
    pkg->xmin = pkg->parms.delx * (imin-1);
    pkg->xmax = pkg->parms.delx * imax;
    pkg->ymin = pkg->parms.dely * (jmin-1);
    pkg->ymax = pkg->parms.dely * jmax;
    
    fseek( f, PARTICLE_START_POS(header), SEEK_SET );
    while( ! feof( f ) )
    {
	fread( &pf, sizeof(pf), 1, f );
	p = (struct Particle *) malloc( sizeof(struct Particle) );
	p->x = pf.x;
	p->y = pf.y;
	if( p == 0 )
	{
	    fprintf( stderr, "Out of memory" );
	    exit( 1 );
	}
	if( pf.particle )
	{
	    if( pf.inject )
	    {
		p->next = pkg->particles[pf.flavor].injection;
		pkg->particles[pf.flavor].injection = p;
	    } else {
		p->next = pkg->particles[pf.flavor].first;
		pkg->particles[pf.flavor].first = p;
	    }
	} else {
	    p->next = pkg->fluids[pf.flavor].first;
	    pkg->fluids[pf.flavor].first = p;
	}
    }
}

void write_DataPackage_to_file
( 
    struct DataPackage *pkg, 
    FILE *f,     
    int write_psiZeta 
)
{
    struct FileHeader header;
    struct ParticleInFile pf;
    struct Particle *p;
    int i;
    
    if  /* Das ist ein vollstaendiges Packet */
    ( 
	pkg->imin == 1 && 
	pkg->jmin == 1 && 
	pkg->imax == floor( pkg->parms.xlength / pkg->parms.delx + 0.5 ) &&
	pkg->jmax == floor( pkg->parms.ylength / pkg->parms.dely + 0.5 )
    )
    {
	header.parms = pkg->parms;
	header.imax  = pkg->imax;
	header.jmax  = pkg->jmax;
	header.t     = pkg->time;
	header.delt  = pkg->delt;
	header.res   = pkg->res;
	header.it    = pkg->it;
	
	fseek( f, FILE_HEADER_POS, SEEK_SET );
	fwrite( &header, sizeof(header), 1, f );
    } else {
	fseek( f, FILE_HEADER_POS, SEEK_SET );
	fread( &header, sizeof(header), 1, f );
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, U_POS(i, pkg->jmin, header), SEEK_SET );
 	fwrite( pkg->U[i]+pkg->jmin, sizeof(REAL), pkg->jmax-pkg->jmin+1, f );
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, V_POS(i, pkg->jmin, header), SEEK_SET );
 	fwrite( pkg->V[i]+pkg->jmin, sizeof(REAL), pkg->jmax-pkg->jmin+1, f );
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, P_POS(i, pkg->jmin, header), SEEK_SET );
 	fwrite( pkg->P[i]+pkg->jmin, sizeof(REAL), pkg->jmax-pkg->jmin+1, f );
    }
    if( write_psiZeta )
    {
	for( i = pkg->imin; i <= pkg->imax; i++ )
	{
	    fseek( f, PSI_POS(i, pkg->jmin, header), SEEK_SET );
	    fwrite( pkg->PSI[i]+pkg->jmin, sizeof(REAL), pkg->jmax-pkg->jmin+1, f );
	}
	for( i = pkg->imin; i <= pkg->imax; i++ )
	{
	    fseek( f, ZETA_POS(i, pkg->jmin, header), SEEK_SET );
	    fwrite( pkg->ZETA[i]+pkg->jmin, sizeof(REAL), pkg->jmax-pkg->jmin+1, f );
	}	
    }
    for( i = pkg->imin; i <= pkg->imax; i++ )
    {
	fseek( f, FLAG_POS(i, pkg->jmin, header), SEEK_SET );
 	fwrite( pkg->FLAG[i]+pkg->jmin, sizeof(int), pkg->jmax-pkg->jmin+1, f );
    }

    fseek( f, PARTICLE_POS( header, pkg->particle_offset ), SEEK_SET );

    pf.particle = TRUE;
    for( i = 0; i < PARTICLE_FLAVORS; i++ )
    {
	pf.flavor = i;
	pf.inject = FALSE;
	for( p = pkg->particles[i].first; p != 0; p = p->next )
	{	    
	    pf.x = p->x;
	    pf.y = p->y;
	    fwrite( &pf, sizeof(struct ParticleInFile), 1, f );
	}
	pf.inject = TRUE;
	for( p = pkg->particles[i].injection; p != 0; p = p->next )
	{
	    pf.x = p->x;
	    pf.y = p->y;
	    fwrite( &pf, sizeof(struct ParticleInFile), 1, f );	    
	}
    }
    pf.particle = FALSE;
    pf.inject = FALSE;
    for( i = 0; i < FLUID_FLAVORS; i++ )
    {
	pf.flavor = i;
	for( p = pkg->fluids[i].first; p != 0; p = p->next )
	{	    
	    pf.x = p->x;
	    pf.y = p->y;
	    fwrite( &pf, sizeof(struct ParticleInFile), 1, f );
	}
    }
}
