/* comm.c */

#include "init.h"
#include "model.h"

#include <sys/time.h>
#include <unistd.h>

struct MasterPackage *ms_pkg;

int answer_received = TRUE;

void set_time( REAL time )
{
    int msg = FDCMPTime;

    model.time = time;
    
    write( ms_pkg->s_interface, &msg, sizeof(msg) );        
    write( ms_pkg->s_interface, &time, sizeof(time) );
}

void set_state( enum CalculationState state )
{
    int msg = FDCMPState;
    
    model.state = state;
    
    write( ms_pkg->s_interface, &msg, sizeof(msg) );        
    write( ms_pkg->s_interface, &state, sizeof(state) );
}

void set_progres( float ratio )
{
    int msg = FDCMPProgres;
    
    model.progres = ratio;
    
    write( ms_pkg->s_interface, &msg, sizeof(msg) );        
    write( ms_pkg->s_interface, &ratio, sizeof(ratio) );
}

void set_residuum( REAL res )
{
    int msg = FDCMPResiduum;
    
    model.res = res;
    
    write( ms_pkg->s_interface, &msg, sizeof(msg) );        
    write( ms_pkg->s_interface, &res, sizeof(res) );
}

void answer_new_file_choices( struct MasterPackage *pkg )
{
    int i;
    int num_choices = NUM_INIT_CHOICES;
    
    write( pkg->s_interface, &num_choices, sizeof(num_choices) );
    
    for( i = 0; i < NUM_INIT_CHOICES; i++ )
	write
	( 
	    pkg->s_interface, 
	    init_choices[i], 
	    FDCMP_NEW_FILE_CHOICE_NAMELENGTH 
	);
}

extern int choice;

struct MasterPackage floatator_main_loop( struct MasterPackage *pkg )
{
    int s = pkg->s_interface;
    fd_set mask;
    struct timeval timeout;
    int ready, msg, success;
    int calculating = FALSE;
    int i, iParm;
    int step = 0;
    struct Particle *p;
    FILE *f;
    
    for(;;)
    {
	if( calculating )
	{
	    poll:
	    /* Poll the communication line */
	    timeout.tv_sec = 0;
	    timeout.tv_usec = 0;
	    FD_ZERO( &mask );
	    FD_SET( s, &mask );
	    ready = select( FD_SETSIZE, &mask, 0, 0, &timeout );
	    if( ready == 0 ) 
	    {
		calculate( pkg );
		if( step >= 5 && answer_received )
		{
		    if( model.output_filename != 0 )
		        free( model.output_filename );
	    	    model.output_filename = tempnam( ".", "flt" );
		    f = fopen( model.output_filename, "w" );
	    	    write_DataPackage_to_file( &model, f, TRUE );
		    fclose( f );
	    	    msg = FDCMPDataSaved;
		    write( s, &msg, sizeof(msg) );
		    msg = strlen( model.output_filename );
		    write( s, &msg, sizeof(msg) );
		    write( s, model.output_filename, msg );
		    answer_received = FALSE;
		    step = -1;
		}
		step++;
		goto poll;
	    }
	}
	success = read( s, &msg, sizeof(msg) );
	if( success != sizeof(msg) )
	{
	    fprintf( stderr, "- Error: to msg is only %d bytes read.\n", success );
	    fprintf( stderr, "-   msg = %d\n", msg );
	    exit( 1 );
	}
	switch( msg )
	{
	  case FDCMPNewFileChoices:
	    answer_new_file_choices( pkg );
	  Case FDCMPNewFile:
	    read( s, &choice, sizeof(choice) );
	    init( pkg, iParm );
	    f = fopen( model.output_filename, "w" );
	    write_DataPackage_to_file( &model, f, TRUE );
	    fclose( f );
	    msg = FDCMPDataSaved;
	    write( s, &msg, sizeof(msg) );
	    msg = strlen( model.output_filename );
	    write( s, &msg, sizeof(msg) );
	    write( s, model.output_filename, msg );
	    answer_received = FALSE;
	  Case FDCMPSaveFile:
	  Case FDCMPQuit:
	    pkg->running = FALSE;
	    return *pkg;
	  Case FDCMPChangeConfig:
	  Case FDCMPChangeParms:
	  Case FDCMPSetUVP:
	  Case FDCMPEmptyArea:
	  Case FDCMPNewObstacle:
	  Case FDCMPNewFluid:
	  Case FDCMPNewParticle:
	    p = (struct Particle *)malloc( sizeof(struct Particle) );
	    if( p == 0 )
	    {
	        fprintf( stderr, "Out of memory\n" );
		exit( 1 );
	    }
	    read( s, &iParm, sizeof(iParm) );
	    read( s, &p->x, sizeof(p->x) );
	    read( s, &p->y, sizeof(p->y) );
	    p->next = model.particles[iParm].first;
	    model.particles[iParm].first = p;
	  Case FDCMPDataRead:
	    answer_received = TRUE;
	  Case FDCMPStartCalc:
	    calculating = TRUE;
	  Case FDCMPStopCalc:
	    calculating = FALSE;
	}
    }	
}
