
#include "init.h"
#include "model.h"

int choice;

char *init_choices[] = {
    "leer", 
    "driven cavity", 
    "Stufe", 
    "Karman'sche Wirbelstrasse", 
    "brechender Damm", 
    "fallender Topfen" 
};

static void enter_particle
( 
    REAL x, REAL y, 
    struct ParticleLine *pl 
)
{
    struct Particle *p;
    p = (struct Particle *) malloc( sizeof(struct Particle) );
    if( p == 0 )
    {
	fprintf( stderr, "Out of memory.\n" );
	exit( 1 );
    }
    p->x = x;
    p->y = y;
    p->next = pl->first;
    pl->first = p;
}


void init_leer( struct MasterPackage *pkg )
{
    
    model.imin = model.jmin = 1;
    model.imax = model.jmax = 10;

    alloc_if_necessary( &model, FALSE, FALSE );
    
    model.xmin = model.ymin = 0.0;
    model.xmax = model.ymax = 1.0;
    model.parms.iproc = model.parms.jproc = 1;
    model.parms.xlength = model.parms.ylength = 1.0;
    model.parms.delx = model.parms.dely = 0.1;
    model.parms.tau = 0.5;
    model.parms.del_vec = 0.0;
    model.parms.itermax = 100;
    model.parms.eps = 0.001;
    model.parms.omg = 1.7;
    model.parms.alpha = 0.9;
    model.parms.Re = 1.0;
    model.parms.GX = model.parms.GY = 0.0;
    model.parms.border[DR_EAST].w = BC_HAFT;
    model.parms.border[DR_EAST].order = 1;
    model.parms.border[DR_EAST].umax = 0.0;
    model.parms.border[DR_EAST].udr = 0.0;
    model.parms.border[DR_NORTH].w = BC_HAFT;
    model.parms.border[DR_NORTH].order = 1;
    model.parms.border[DR_NORTH].umax = 0.0;
    model.parms.border[DR_NORTH].udr = 0.0;
    model.parms.border[DR_WEST].w = BC_HAFT;
    model.parms.border[DR_WEST].order = 1;
    model.parms.border[DR_WEST].umax = 0.0;
    model.parms.border[DR_WEST].udr = 0.0;
    model.parms.border[DR_SOUTH].w = BC_HAFT;
    model.parms.border[DR_SOUTH].order = 1;
    model.parms.border[DR_SOUTH].umax = 0.0;
    model.parms.border[DR_SOUTH].udr = 0.0;
    model.parms.delt_trace = 0.1;
    model.parms.delt_inject = 1.0;
	    		    
    model.output_filename = tempnam( ".", "flt" );
    model.time = 0.0;
}

void init_dcavity(struct MasterPackage *pkg )
{
    
}

void init_stufe(struct MasterPackage *pkg )
{
    
}

void init_karman(struct MasterPackage *pkg )
{
    
}

void init_damm(struct MasterPackage *pkg )
{
    
}

void init_tropfen(struct MasterPackage *pkg )
{
    int i, j;
    REAL stepx, stepy;
    REAL x, y;
    REAL cx, cy, r;
    
    model.imin = model.jmin = 1;
    model.imax = 40;
    model.jmax = 30;

    alloc_if_necessary( &model, TRUE, TRUE );
    
    model.xmin = model.ymin = 0.0;
    model.xmax = 8.0;
    model.ymax = 6.0;
    model.parms.iproc = model.parms.jproc = 1;
    model.parms.xlength = 8.0;
    model.parms.ylength = 6.0;
    model.parms.delx = model.parms.dely = 0.2;
    model.parms.tau = 0.2;
    model.parms.del_vec = 0.5;
    model.parms.itermax = 500;
    model.parms.eps = 0.001;
    model.parms.omg = 1.7;
    model.parms.alpha = 0.5;
    model.parms.Re = 40.0;
    model.parms.GX = 0.0;
    model.parms.GY = -1.0;
    model.parms.border[DR_EAST].w = BC_RUTSCH;
    model.parms.border[DR_EAST].order = 1;
    model.parms.border[DR_EAST].umax = 0.0;
    model.parms.border[DR_EAST].udr = 0.0;
    model.parms.border[DR_NORTH].w = BC_AUS;
    model.parms.border[DR_NORTH].order = 1;
    model.parms.border[DR_NORTH].umax = 0.0;
    model.parms.border[DR_NORTH].udr = 0.0;
    model.parms.border[DR_WEST].w = BC_RUTSCH;
    model.parms.border[DR_WEST].order = 1;
    model.parms.border[DR_WEST].umax = 0.0;
    model.parms.border[DR_WEST].udr = 0.0;
    model.parms.border[DR_SOUTH].w = BC_RUTSCH;
    model.parms.border[DR_SOUTH].order = 1;
    model.parms.border[DR_SOUTH].umax = 0.0;
    model.parms.border[DR_SOUTH].udr = 0.0;
    model.parms.delt_trace = 0.1;
    model.parms.delt_inject = 1.0;
    for( i = model.imin; i <= model.imax; i++ )
	for( j = model.jmin; j <= model.jmax; j++ )
	{
	    model.FLAG[i][j] = B_C;
	    model.U[i][j] = 0.0;
	    model.V[i][j] = 0.0;
	    model.P[i][j] = 0.0;
	}
    for( i = 0; i < PARTICLE_FLAVORS; i++ )
    {
	model.particles[i].first = 0;
	model.particles[i].injection = 0;
    }
    for( i = 0; i < FLUID_FLAVORS; i++ )
    {
	model.fluids[i].first = 0;
	model.fluids[i].injection = 0;
    }
    
    model.output_filename = tempnam( ".", "flt" );
    model.time = 0.0;
    
    for( i = model.imin; i <= model.imax; i++ )
	for( j = model.jmin; j <= model.jmax; j++ )
	{
	    model.FLAG[i][j] = 0;
	    model.U[i][j] = 0.0;
	    model.V[i][j] = 0.0;
	    model.P[i][j] = 0.0;
	}
    for( i = 0; i < PARTICLE_FLAVORS; i++ )
    {
	model.particles[i].first = 0;
	model.particles[i].injection = 0;
    }
    for( i = 0; i < FLUID_FLAVORS; i++ )
    {
	model.fluids[i].first = 0;
	model.fluids[i].injection = 0;
    }
    
    stepx = model.parms.delx / 4;
    stepy = model.parms.dely / 4;
    cx = model.parms.xlength / 2;
    cy = model.parms.ylength / 1.5;
    r = model.parms.ylength / 10;
    for( x = cx-r + stepx/2; x < cx+r; x += stepx )
	for( y = cy-r + stepy/2; y < cy+r; y += stepy )
	    if( SQR(x-cx) + SQR(y-cy) <= SQR(r) )
		enter_particle( x, y, &model.fluids[0] );
    for( x = cx-r; x <= cx+r; x += model.parms.delx )
	for( y = cy-r; y <= cy+r; y += model.parms.dely )
	{
	    int i = (int)(x / model.parms.delx) + 1;
	    int j = (int)(y / model.parms.dely) + 1;
	    model.V[i][j] = -2.0;
	}    
    for( x = stepx/2; x < model.parms.xlength; x += stepx )
	for( y = stepy/2; y < model.parms.ylength/3; y += stepy )
	    enter_particle( x, y, &model.fluids[1] );
}

extern void init_compatible();

void init( struct MasterPackage *pkg, int choice )
{
#if 0
    switch( choice )
    {
      case 0:
	init_leer( pkg );
      Case 1:
        init_dcavity( pkg );
      Case 2:
	init_stufe( pkg );
      Case 3:
	init_karman( pkg );
      Case 4:
        init_damm( pkg );
      Case 5:
	init_tropfen( pkg );
    }   
#else
    init_compatible();
    if( model.output_filename != 0 )
	free( model.output_filename );
    model.output_filename = tempnam( ".", "flt" );
    model.time = 0.0;
#endif
}
