/* init.c */

#include "flag.h"
#include "init2.h"

#include <assert.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define min(x,y) ((x)<(y)?(x):(y))

static char scan_string[] = 
"imax = %d	jmax = %d	xlength = %lf	ylength = %lf\n\
delt = %lf	Tfin = %lf	tau = %lf	del_vec = %lf\n\
eps = %lf	omg = %lf	alpha = %lf	itermax = %d\n\
GX = %lf	GY = %lf	Re = %lf\n\
UI = %lf	VI = %lf	PI = %lf        ppc = %d\n\
Problem = %s		        Visualisierung = %s\n\
wl = %d		wr = %d		wt = %d		wb = %d\n\
N = %d\n\
x1 = %lf	y1 = %lf	x2 = %lf	y2 = %lf\n\
delt_trace = %lf\n\
delt_inject = %lf\n\
delt_streak = %lf\n";

int READ_PARAMETER
(
    char *Inputfile,
    REAL *xlength, REAL *ylength,
    int *imax, int *jmax,
    REAL *delx, REAL *dely,
    REAL *delt, REAL *t_end, 
    REAL *tau, int *itermax, 
    REAL *eps, REAL *omg, REAL *alpha, REAL *Re,
    REAL *GX, REAL *GY, REAL *UI, REAL *VI, REAL *PI,
    REAL *del_vec,
    char *problem, int *wl, int *wr, int *wt, int *wb,
    char *visualisierung,  
    int *N, REAL *x1, REAL *y1, REAL *x2, REAL *y2, 
    REAL *delt_trace, REAL *delt_inject, REAL *delt_streak,
    int *ppc
)
{
    FILE * f = fopen( Inputfile, "r" );
    if( f == 0 ) return 0;
    if( fscanf( f, scan_string,
        imax, jmax, xlength, ylength,
        delt, t_end, tau, del_vec,
        eps, omg, alpha, itermax,
        GX, GY, Re,
        UI, VI, PI, ppc,
        problem, visualisierung, 
        wl, wr, wt, wb, 
	N, x1, y1, x2, y2, 
	delt_trace, delt_inject, delt_streak
    ) == 0 ) return 0;
    printf( "Parameter '%s' eingelesen:\n", Inputfile );
    printf( scan_string,
        *imax, *jmax, *xlength, *ylength,
        *delt, *t_end, *tau, *del_vec,
        *eps, *omg, *alpha, *itermax,
        *GX, *GY, *Re,
        *UI, *VI, *PI, *ppc,
        problem, visualisierung, 
        *wl, *wr, *wt, *wb, 
	*N, *x1, *y1, *x2, *y2, 
	*delt_trace, *delt_inject, *delt_streak
    ); 
    *delx = *xlength / *imax;
    *dely = *ylength / *jmax;
    fclose( f );
    return -1;
}

REAL **RMATRIX( int nrl, int nrh, int ncl, int nch )
{
    int i;
    REAL **m;
    if( (m=(REAL**)malloc((unsigned)(nrh-nrl+1)*sizeof(double*)))==NULL )
    {
        printf( "kein Speicher mehr da.\n" );
        exit( 1 );
    }
    m -= nrl;
    for( i=nrl; i<=nrh; i++ )
    {
        if( (m[i]=(REAL*)malloc((unsigned)(nch-ncl+1)*sizeof(double)))==NULL )
        {
            printf( "kein Speicher mehr da.\n" );
            exit( 1 );
        }
        memset( m[i], 0, (unsigned)(nch-ncl+1)*sizeof(double) );
        m[i] -= ncl;
    }
    return m;
}

void FREE_MATRIX( REAL **m, int nrl, int nrh, int ncl, int nch )
{
    int i;
    for( i=nrh; i >nrl; i-- ) free( (char*) (m[i]+ncl));
    free( (char*) (m+nrl) );
}

void INIT_UVP
( 
    REAL **U, REAL **V, REAL **P, 
    int imax, int jmax, 
    REAL UI, REAL VI, REAL PI, 
    char *problem
)
{
    int i,j;
    for( i = 0; i <= imax+1; i++ )
        for( j = 0; j <= jmax+1; j++ ) 
        {
            U[i][j] = UI;
            V[i][j] = VI;
            P[i][j] = PI;
        }
    if( strcasecmp( "Stufe",  problem ) == 0 )
    {
	for( i = 0; i <= imax+1; i++ )
	    for( j = 0; j <= (jmax+1)/2; j++ ) 
	    {
		U[i][j] = 0.0;
	    }
	    
    }
}

int **IMATRIX( int nrl, int nrh, int ncl, int nch )
{
    int i;
    int **m;
    if( (m=(int**)malloc((unsigned)(nrh-nrl+1)*sizeof(int*)))==NULL )
    {
        printf( "kein Speicher mehr da.\n" );
        exit( 1 );
    }
    m -= nrl;
    for( i=nrl; i<=nrh; i++ )
    {
        if( (m[i]=(int*)malloc((unsigned)(nch-ncl+1)*sizeof(int)))==NULL )
        {
            printf( "kein Speicher mehr da.\n" );
            exit( 1 );
        }
        memset( m[i], 0, (unsigned)(nch-ncl+1)*sizeof(int) );
        m[i] -= ncl;
    }
    return m;
}

void INIT_FLAG( char *problem, int **FLAG, int imax, int jmax )
{
    int i, j;
    for( i = 1; i <= imax; i++ )
	for( j = 1; j <= jmax; j++ )
	    FLAG[i][j] = C_F;
    if( strcasecmp( "DCavity", problem ) == 0 )
    {
       /* do nothing */;
    }
    else if( strcasecmp( "Karman", problem ) == 0 )
    {
        int l = jmax / 5;
	int width = 3;
	int iStart = 2 * l;
	int jStart = iStart;
	int iEnd = iStart + l;
	int jEnd = jStart + l;
	int m, n;
	/* Balken plotten */
	for( n = 0; n < l; n++ )
	{
	    for( m = -width/2; m <= width/2; m++ )
	        FLAG[iStart+n+m][jStart+n-m] = C_B;
	    for( m = -width/2+1; m <= width/2; m++ )
		FLAG[iStart+n+m][jStart+n-m+1] = C_B;
	}
	/* Ecken abrunden */
	for( n = -width/2; n <= width/2; n++ )
	    for( m = -width/2; m <= width/2; m++ )
	        FLAG[n+iStart][m+jStart] = C_B;
	for( n = -width/2; n <= width/2; n++ )
	    for( m = -width/2; m <= width/2; m++ )
	        FLAG[n+iEnd][m+jEnd] = C_B;
    }
    else if( strcasecmp( "Stufe", problem ) == 0 )
    {
        int l = jmax / 2;
	for( i = 1; i <= min( l, imax ); i++ )
	    for( j = 1; j <= l; j++ )
	        FLAG[i][j] = C_B;
    }
    printf( "Erzeugte Geometrie:\n" );
    printf( "/" );
    for( i = 1; i <= imax; i++ )
        printf( "-" );
    printf( "\\\n" );
    for( j = jmax; j >= 1; j-- )
    {
        printf( "|" );
        for( i = 1; i <= imax; i++ )
	    printf( "%c", FLAG[i][j] == C_F ? ' ' : '*' );
	printf( "|\n" );
    }
    printf( "\\" );
    for( i = 1; i <= imax; i++ )
        printf( "-" );
    printf( "/\n" );
    for( i = 1; i <= imax; i++ )
	for( j = 1; j <= jmax; j++ )
	{
	    if( FLUID(FLAG[i][j+1]) )
	        FLAG[i][j] |= B_N;
	    if( FLUID(FLAG[i][j-1]) )
	        FLAG[i][j] |= B_S;
	    if( FLUID(FLAG[i-1][j]) )
	        FLAG[i][j] |= B_W;
	    if( FLUID(FLAG[i+1][j]) )
	        FLAG[i][j] |= B_O;
	    assert( ! VERBOT(FLAG[i][j]) );
	}
}
