/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      model.h                                                *
*                                                                    *
*  Contents:  The definition of the problem specific part            *
*             of the floatator program.                              *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  11.07.95 First steps                             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

#ifndef MODEL_H
#define MODEL_H

#include "../glib.h"

/********************************************************************\
*                                                                    *
*                      Definition von REAL                           *
*                      ===================                           *
*                                                                    *
*  Comment: Hier kann die genauigkeit der Berechnung eingestellt     *
*           werden.                                                  *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon   Definition von REAL = double                   *
*                                                                    *
\********************************************************************/

/* Definition des Typs */
typedef double REAL;
/* String der an printf/scanf uebergeben werden muss, um ein */
/* REAL zu scannen                                           */ 
#define REAL_FSTRING "lf"

/********************************************************************\
*                                                                    *
*                           Definitionen                             *
*                           ============                             *
*                                                                    *
*  Comment: In alphabetischer Reihenfolge                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  05.09.95  Gordon  Definition von der Fallbacks fuer               *
*                    Socket und Interface Namen.                     *
*                                                                    *
\********************************************************************/

#define SOCKET_NAME    "floatator.socket"
#define INTERFACE_NAME "floatator_X_interface"

/********************************************************************\
*                                                                    *
*                         Enum-Konstanten                            *
*                         ===============                            *
*                                                                    *
*  Comment: In alphabetischer Reihenfolge                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von BorderCondition, Direction,      *
*                    und Flag                                        *
*                                                                    *
\********************************************************************/

/* Bedingungen fuer die Raender des Simulationsgebietes */
enum BorderCondition
{
    BC_EIN,     /* Einstroembedingung        */
    BC_HAFT,    /* Haftfbedingung            */
    BC_RUTSCH,  /* Rutschbedingung           */
    BC_AUS,     /* Ausstroembedingung        */
    BC_DC       /* "Driven-Cavity"-Bedingung */
};

/* Verschiede Zustaende waehrend der Berechnung */
enum CalculationState
{
    CS_NONE,      /* kein gueltiger Zustand                       */
    CS_MARK,	  /* Markierung der Fluidzellen		          */
    CS_BCOND,	  /* Markierung der Oberflaechenzellen des Fluids */
    CS_SOR,	  /* SOR-Iteration                                */
    CS_FLUIDS,    /* Verschieben der Fluid-Gebiete                */
    CS_PARTICLES, /* Bewegen der Partikel                         */  
    CS_NUMSTATES  /* Dummy, Anzahl der Zustaende                  */
};

/* Konstanten fuer die Himmelsrichtungen */
enum Direction
{
    DR_EAST,
    DR_NORTH,
    DR_WEST,
    DR_SOUTH,
    DR_ALL
};

/* Flags der Zellen */
enum Flag
{
    B_N = 1,    /* Die noerdliche Zelle ist kein Hindernis */
    B_S = 2,    /* Die suedliche Zelle ist kein Hindernis  */
    B_W = 4,    /* Die westliche Zelle ist kein Hindernis  */
    B_E = 8,    /* Die suedliche Zelle ist kein Hindernis  */
    B_C = 16,   /* Die Zelle selbst ist kein Hindernis     */

    C_N = 32,   /* Die noerdliche Zelle ist leer           */
    C_S = 64,   /* Die suedliche Zelle ist leer            */
    C_W = 128,  /* Die westliche Zelle ist leer            */
    C_E = 256,  /* Die oestliche Zelle ist leer            */
    C_C = 512,  /* Die Zelle selbst ist leer               */

    /* Masken fuer die Abfrage bewstimmter Eigenschaften */
    B_ALL = B_N + B_S + B_W + B_E + B_C,
    C_ALL = C_N + C_S + C_W + C_E + C_C,

    /* Zusammengesetzte Rand-Eigenschaften */
    B_NE = B_N + B_E,
    B_SE = B_S + B_E,
    B_SW = B_S + B_W,
    B_NW = B_N + B_W,

    /* Zusammengesetzte Fluid-Eigenschaften */
    C_NE = C_N + C_E,
    C_NW = C_N + C_W,
    C_SE = C_S + C_E,
    C_SW = C_S + C_W,
    C_NS = C_N + C_S,
    C_WE = C_W + C_E,
    C_NSWE = C_N + C_S + C_W + C_E,
    C_NSW = C_N + C_S + C_W,
    C_NSE = C_N + C_S + C_E,
    C_NWE = C_N + C_W + C_E,
    C_SWE = C_S + C_W + C_E
};

/* Modus der Partikel */
enum ParticleMode
{
    PM_TRACING,          /* Bahnverfolgung */
    PM_STREAKLINES       /* Streichlinien  */
};

/********************************************************************\
*                                                                    *
*                         Datenstrukturen                            *
*                         ===============                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von BorderSpec, CalcParms, Particle, *
*                     ParticleLine, DataPackage, SlavePackage,       *
*                     SlaveInfo und MasterPackage                    *
*                    Definition der Makros FLUID_FLAVORS und         *
*                     PARTICLE_FLAVORS                               *
*                                                                    *
\********************************************************************/

/* Eigenschaften der Gebiertsraender */
struct BorderSpec
{
    enum BorderCondition w; /* um welche Randart handelt es sich ? */
    int  order;             /* Ordnung und Geschwindigkeit eines   */
    REAL umax;              /* Einstroemprofiles                   */
    REAL udr;               /* Geschwindigkeit fuer "DCavity"      */
};

/* Parameter, die die Berechnung betreffen */
struct CalcParms
{
    int iproc, jproc;       /* Anzahl der Prozessoren                   */
    REAL xlength, ylength;  /* Breite und Hoehe des Gebietes            */
    REAL delx, dely;        /* Diskretisierungsfeinheit                 */
    REAL tau;               /* Sicherheitsfaktor der Zeitschrittweiten- */
                            /* steuerung.                               */
                            /* 0 < tau <= 1                             */
    REAL del_vec;           /* Frame-rate der Ausgabe                   */
    int  itermax;           /* Maximalzahl der SOR-Schritte             */
    REAL eps;               /* Genauigkeitsgrenze der SOR-Iteration     */
    REAL omg;               /* Relaxationsfaktor der SOR-Iteration      */
                            /* 0 < omg <= 2                             */
    REAL alpha;             /* Upwind-Differencing-Factor               */
                            /* 0 <= alpha <= 1                          */
                            /* alpha = 0: Verfahren 2. Ordnung          */
                            /* alpha = 1: Donor-Cell-Schema             */
    REAL Re;                /* Reynoldszahl                             */
    REAL GX, GY;            /* Gravitation                              */
    struct BorderSpec border[DR_ALL];
                            /* Eigenschaften der Gebietsraender         */
    REAL delt_trace;        /* Zeitspanne, nach der die Partikel bewegt */
                            /* werden.                                  */
    REAL delt_inject;       /* Zeitspanne, nach der neue Partikel       */
                            /* injeziert werden.                        */
};

/* Anzahl der verschiedenen Fluessigkeits-Typen */
#define FLUID_FLAVORS 5
/* Anzahl der verschiedenen Partikel-Typen */
#define PARTICLE_FLAVORS 6

struct Particle
{
    REAL x, y;             /* Position des Partikels */
    struct Particle *next; /* verkettete Liste       */
};

struct ParticleLine
{
    struct Particle *first;     /* verkettete Liste der Partikel */
    struct Particle *injection; /* verkettete Liste, wo Partikel */
                                /* injeziert werden.             */ 
};

struct DataPackage
{
    /* Ausmasse der gespeicherten Arrays */ 
    int imin, imax, jmin, jmax;
    /* Ausmasse des gespeicherten Gebietes */
    REAL xmin, xmax, ymin, ymax;

    /* Parameter der Berechnung */
    struct CalcParms parms;

    /* Zustand des Gebietes */
    int  **FLAG; /* Gebietskonfiguertion */
    REAL **U;    /* Horizontalgeschwindigkeit des Fluids */
    REAL **V;    /* Vertikalgeschwindigkeit des Fluids */
    REAL **P;    /* Druck des Fluids */

    /* Hilfsgroessen waehrend der Berechnung */
    REAL **F, **G; /* Geschindigkeiten der disketen Impulsgleichunen */
    REAL **RHS;    /* Rechte Seite der Poisson-Iteration             */

    /* Hilfsgroessen zur Visualisierung */
    REAL **PSI;    /* Stromfunktion */
    REAL **ZETA;   /* wirbelfunktion */
       
    /* Partikel */
    struct ParticleLine particles[PARTICLE_FLAVORS]; 
    struct ParticleLine fluids[FLUID_FLAVORS];
    int particle_offset;  /* Offset, an dem Partikel in die Aufgabe- */
                          /* Datei geschrieben werden duerfen.       */

    /* Name der Ausgabedatei */
    char *output_filename;

    /* sonstiges */
    REAL time;	    /* Aktuelle Zeit                    */
    REAL delt;      /* Zeitschrittweite der Berechnung  */
    enum CalculationState state;
	            /* aktueller Zustand der Berechnung */
    float progres;  /* Fortschritt der Berechnung in %  */
    REAL res;       /* Residuum der Iteration           */
    int  it;        /* Anzahl der Iterationsschritte    */ 
};

struct SlavePackage
{
    int mytid;              /* Task-ID des Slaves    */
    int master;             /* Task-ID des Masters   */
    int neighbours[DR_ALL]; /* Task-IDs der Nachbarn */
};

struct SlaveInfo
{
    /* Ausmasse des zustaendigen Gebietes */
    REAL xmin, xmax, ymin, ymax;
    /* Ausmasse der gespeicherten Arrays */
    int imin, imax, jmin, jmax;
    int tid;                /* Task-ID des Slaves    */
    /* Partikel, die sich im naechsten Zeitschritt im Gebiet */
    /* des Slave befinden.                                   */
    struct ParticleLine particles[PARTICLE_FLAVORS]; 
    struct ParticleLine fluids[FLUID_FLAVORS];
    int num_particles;      /* Anzahl der Partikel-slots, die ein    */
                            /* Slave in der Ausgabe-Datei benoetigt. */
    int particle_offset;    /* Offset, an dem ein Slave seine Partikel */
                            /* in die Ausgabe-Datei schreiben darf     */ 
};

struct MasterPackage
{
    /* Kommunikation mit den Slaves */
    int mytid;                /* Task-ID des Masterns       */
    struct SlaveInfo **slave; /* Array der Slaves           */

    /* Kommunikation mit dem User-Interface */
    int s_interface;          /* Socket des User-Interfaces        */
    char *socket_name;        /* Name und evtl. Pfad des Socket    */
    int pid_interface;        /* Process-ID des User-Interface     */
    char *interface_name;     /* Name und evtl. Pfad des Interface */

    /* sonstiges */
    REAL t;                   /* aktuelle, absolute Zeit              */ 
    int running;	      /* Soll das Programm noch weiter laufen */
};

/********************************************************************\
*                                                                    *
*                      Ausgabe-Datei-Format                          *
*                      ====================                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  12.07.95  Gordon   First steps                                    *
*                                                                    *
*                                                                    *
*                                                                    *
\********************************************************************/

/* Mithilfe des Pads kann in die Datei noch der Zustand des */
/* Visualisierungsprogramms mit aufgenomment werden.        */
#define FILE_PAD_SIZE 0

struct FileHeader
{
    struct CalcParms parms; /* Eingangsparameter der Berechnung */
    int imax, jmax;         /* Ausmasse des UVP-Arrays          */

    /* Ausgabeparameter der Berechnung */
    REAL t;                 /* aktuelle, absolute Zeit          */ 
    REAL delt;              /* Zeitschrittweite der Berechnung  */
    REAL res;               /* Residuum der Iteration           */
    int  it;                /* Anzahl der Iterationsschritte    */
};

/* Da verkettete Listen in Dateien eine zentrale Speicherverwaltung */
/* benoetigten, werden die Partikel in der Ausgabe-Datei flach,     */
/* d.h. mit redundanter Zusatzinformation abgespeichert.            */ 
struct ParticleInFile
{
    int  particle :1;   /* FALSE: Fluid-Partikel               */
                        /* TRUE:  Partikel                     */
    int  inject   :1;   /* FALSE: normaler Partikel            */
                        /* TRUE:  Partikel in der inject-Liste */
    int  flavor   :6;   /* In welchen Typ hat der Partikel     */
    REAL x, y;          /* Position des Partikels              */
};

/* Hilfsfunktion zur Berechnung eines Indexes in der Datei */
#define CTGIX(i,j, header, n)					      \
                           (sizeof(struct FileHeader) + FILE_PAD_SIZE \
                            + ( (header).imax * (header).jmax )       \
                              * sizeof(REAL) * (n)                    \
                            + ( ((i)-1) * (header).jmax + ((j)-1) )   \
                              * sizeof(REAL)                          \
                           )

/* Funktionen zur Bestimmung der Position der einzelnen */
/* Komponenten in der Datei.                            */
#define FILE_HEADER_POS     0
#define U_POS(i,j, header)          CTGIX(i,j, header, 0)
#define V_POS(i,j, header)          CTGIX(i,j, header, 1)
#define P_POS(i,j, header)          CTGIX(i,j, header, 2)
#define PSI_POS(i,j, header)        CTGIX(i,j, header, 3)
#define ZETA_POS(i,j, header)       CTGIX(i,j, header, 4)
#define FLAG_POS(i,j, header)       CTGIX(i,j, header, 5)
#define PARTICLE_START_POS(header)  CTGIX(1,1, header, 6)
#define PARTICLE_POS(header, offset)			    \
    ( PARTICLE_START_POS(header) +			    \
      (offset) * sizeof(struct ParticleInFile) )


/********************************************************************\
*                                                                    *
*                Protokoll zwichen Master und Slave                  *
*                ==================================                  *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  12.07.95  Gordon   First steps                                    *
*                                                                    *
\********************************************************************/

enum MSCommands
{
    MS_Aera,    /* Der Slave wird fuer ein Gebiet eingeteilt     */
    MS_File,    /* Der Dateiname wird festgelegt                 */
    MS_Load,    /* Die Datei wird geladen                        */
    MS_Save,    /* Die Daten werden abgespeichert                */
    MS_Parms    /* Die Berechnungsparameter haben sich geaendert */
};

/* Parameter fuer das MS_Area Kommando */
struct MS_Init_Parms
{
    /* Ausmasse der gespeicherten Arrays */
    int  imin, imax, jmin, jmax;
    /* Ausmasse des gespeicherten Gebietes */
    REAL xmin, xmax, ymin, ymax;
    int  neighbours[DR_ALL];     /* Task-IDs der Nachbarn      */
    char filename[20]; /* Dateiname der Ausgabedatei */ 
};

/********************************************************************\
*                                                                    *
*      FDCMP - Floatator direct communication message protocol       *
*      =======================================================       *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  14.07.95  Gordon   First steps                                    *
*  01.08.95  Gordon   Definition des Protokolls fertig               *
*  28.08.95  Gordon   Hiunzufuegung von Prozeduren zum Senden und    *
*                     Empfangen der Nachrichten bzw. Kommandos.      *
*                                                                    *
\********************************************************************/

/* Kommandos an den Floatator-Master */
enum FDCMPCommands
{
    FDCMPNewFileChoices,   /* Vordefinierte Moeglichkeiten fuer eine     */
                           /* neue Simulation werden erfragt             */
    FDCMPNewFile,          /* Eine neue Simulation soll gestartet werden */
    FDCMPOpenFile,         /* Eine neue Datei soll geladen werden        */
    FDCMPSaveFile,         /* Eine Datei soll gespeichert werden         */
    FDCMPQuit,             /* Der Floatator soll beendet werden          */

    FDCMPChangeConfig,     /* Die Konfiguration muss geaendert werden    */
    FDCMPChangeParms,      /* Die Berechnungsparameter wurden veraendert */

    FDCMPSetUVP,           /* U, V und P sollen in einem Gebiet einheit- */
                           /* lich gesetzt werden                        */
    FDCMPEmptyArea,        /* Ein Gebiet soll auf "leer" gesetzt werden  */
    FDCMPNewObstacle,      /* Ein Hindernis soll gesetzt werden          */
    FDCMPNewFluid,         /* Fluid soll eingebracht werden              */
    FDCMPNewParticle,      /* Neue Partikel sollen eingebracht werden    */

    FDCMPStartCalc,        /* Start der Berechnung                       */
    FDCMPStopCalc,         /* Stop der Berechnung                        */
    FDCMPDataRead          /* Rueckmeldung zum Empfang von DataSaved     */
};

/* Rueckfragen vom Floatator-Master */
enum FDCMPMessages
{
    FDCMPTime,             /* t und delt wurden berechnet                */
    FDCMPState,            /* Der Zustand der Berechnung hat sich geaendert */
    FDCMPProgres,          /* Fortschritt der Berechnung                 */
    FDCMPResiduum,         /* Das Residuum der wurde berechnet           */
    FDCMPDataSaved         /* Die Daten wurden abgespeichert             */
};

/* Parameter fuer die einzelnen Kommandos */

#define FDCMP_NEW_FILE_CHOICE_NAMELENGTH 32

struct FDCMPTimeMessage
{
    enum FDCMPMessages reason; /* FDCMPTime */
    REAL time;  
};

struct FDCMPNewFileChoices_Params
{
    int num_choices;
    struct {
        char name[FDCMP_NEW_FILE_CHOICE_NAMELENGTH];
    } choice[1];
};

struct FDCMPNewFile_Params
{
    int start_configuration;
};

struct FDCMPOpenSaveFile_Params
{
    int name_length;
    char name[1];
};

struct FDCMPAreaManipulation_Params
{
    REAL x1, y1, x2, y2;
    int flavor;
    REAL UI, VI, PI;
};

/********************************************************************\
*                                                                    *
*                        globale Variablen                           *
*                        =================                           *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  30.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

/********************************************************************\
*                                                                    *
*                            Funktionen                              *
*                            ==========                              *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

void model_init();
void model_cleanup();

/* Actions performed by the floatator */
void init_MasterPackage( struct MasterPackage *pkg, int argc, char **argv, char **env );
int create_socket( char *socket_name );
int spawn_interface( char *interface_name, char **argv, char **env );
int accept_from_socket( int s );
struct MasterPackage floatator_main_loop( struct MasterPackage *pkg );
void calculate( struct MasterPackage *pkg );
void purge_socket( struct MasterPackage *pkg );

/* Actions performed by the interface */
int connect_to_socket( char *socket_name );

extern struct DataPackage model;

void write_DataPackage_to_file
( 
    struct DataPackage *pkg, 
    FILE *f,     
    int write_psiZeta 
);
void read_DataPackage_from_file
( 
    struct DataPackage *pkg,
    FILE *f, 
    int read_psiZeta 
);
void alloc_if_necessary
( 
    struct DataPackage *pkg, 
    int alloc_psiZeta, 
    int alloc_fGRhs
);

void set_time( REAL time);
void set_state( enum CalculationState state );
void set_progres( float ratio );
void set_residuum( REAL res );

#endif
