/* surface.c */

#include "flag.h"
#include "surface.h"
#include "visual.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "model_compat.h"

#define SQR(x) ((x)*(x))
#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((x)<(y)?(y):(x))
#define abs(x) ((x)<0?-(x):(x))
#define Case break;case
#define Default break;default

static int num_particles = 0;
static int num_fluids = 0;

#define PARTICLE(x,y) particle((x),(y), &model.fluids[n] ); num_fluids++ 
static void particle( REAL x, REAL y, struct ParticleLine *pl )
{
    struct Particle *p;
    p = (struct Particle *) malloc( sizeof(struct Particle) );
    assert( p );
    p->x = x;
    p->y = y;
    p->next = pl->first;
    pl->first = p;
}

struct particleline *INIT_PARTICLES
(
    int *N,
    int imax, int jmax,
    REAL delx, REAL dely,
    int ppc, char *problem,
    REAL **U, REAL **V
)
{
    int n = 0;
    REAL stepx = delx / ppc;
    REAL stepy = dely / ppc;
    REAL xlength = delx * imax;
    REAL ylength = dely * jmax;
    REAL x, y;
    for( n = 0; n < FLUID_FLAVORS; n++ )
        model.fluids[n].first = 
	model.fluids[n].injection = 0;
    n = 0;
    if( strcasecmp( "FallenderTropfen", problem ) == 0 )
    {
	*N = 2;
	n = 0;
	{
	    REAL cx = xlength/2;
	    REAL cy = ylength/1.5;
	    REAL r = ylength / 10;
	    
	    for( x = cx-r + stepx/2; x < cx+r; x += stepx )
	        for( y = cy-r + stepy/2; y < cy+r; y += stepy )
		    if( SQR(x-cx) + SQR(y-cy) < SQR(r) )
#ifndef NDEBUG
		    {
			int i, j;
			i = (int)(x / delx) + 1;
			j = (int)(y / dely) + 1;
			assert( i >= 1 && i <= imax && j >= 1 && j <= jmax );
#endif
		        PARTICLE( x, y );
#ifndef NDEBUG
		    }
#endif
	    for( x = cx - r; x <= cx + r; x += delx )
	        for( y = cy - r; y <= cy + r; y += dely )
		{
		    int i, j;
		    i = (int)(x / delx) + 1;
		    j = (int)(y / dely) + 1;
		    assert( i >= 1 && i <= imax && j >= 1 && j <= jmax );
		    U[i][j] = 0.0;
		    V[i][j] = -2.0;
		}
		    
	}
	n = 1;
	for( x = stepx/2; x < xlength; x += stepx )
	    for( y = stepy/2; y < ylength/3; y += stepy )
	        PARTICLE( x, y );
    } else if( strcasecmp( "BrechenderDamm", problem ) == 0 )
    {
	*N = 1;
	n = 0;
	for( x = stepx/2; x < xlength/5; x += stepx )
	    for( y = stepy/2; y < ylength; y += stepy )
	        PARTICLE( x, y );
    } else
    {
        *N = 0;
        return 0;
    }
    return 0;
}

void MARK_CELLS
(
    int **FLAG,
    int imax, int jmax,
    REAL delx, REAL dely,
    int N, struct particleline *Partlines   
)
{
    int i,j;
    int n;
    int count = 0;
    struct Particle *p;

    /* Erster Schritt: Alle Flags loeschen */
    for( i = 0; i <= imax+1; i++ )
    {
        FLAG[i][0] &= ~SF;
	FLAG[i][jmax+1] &= ~SF;
    }
    for( j = 0; j <= jmax+1; j++ )
    {
        FLAG[0][j] &= ~SF;
	FLAG[imax+1][j] &= ~SF;
    }
    for( i = 1; i <= imax; i++ )
	for( j = 1; j <= jmax; j++ )
	    if( ! HINDER( FLAG[i][j] ) )
	        FLAG[i][j] |= SF;
    
    /* Zweiter Schritt: Zellen mit Fluid markieren */
    for( n = 0; n < FLUID_FLAVORS; n++ )
    {
	for( p = model.fluids[n].first; p; p = p->next )
	{
	    i = (int)(p->x / delx) + 1;
	    j = (int)(p->y / dely) + 1;
	    assert( i >= 1 && i <= imax && j >= 1 && j <= jmax );
	    FLAG[i][j] &= ~C_L;
	    count ++;
        }
	set_progres( (float) count / (float) num_fluids );
    }

    /* Dritter Schritt: Restliche Flags richtig setzen */ 
    for( i = 1; i <= imax; i++ ) 
	for( j = 1; j <= jmax; j++ ) 
	{
	    if( !LEER(FLAG[i-1][j]) )
		FLAG[i][j] &= ~C_W;
	    if( !LEER(FLAG[i+1][j]) )
		FLAG[i][j] &= ~C_O;
	    if( !LEER(FLAG[i][j-1]) )
		FLAG[i][j] &= ~C_S;
	    if( !LEER(FLAG[i][j+1]) )
		FLAG[i][j] &= ~C_N;
	} 
}

void SET_UVP_SURFACE
(
    REAL **U, REAL **V,
    REAL **P, int **FLAG,
    int imax, int jmax, 
    REAL Re, 
    REAL delx, REAL dely, REAL delt,
    REAL GX, REAL GY
)
{
    int i, j;

    for( i = 1; i <= imax; i++ )
	for( j = 1; j <= jmax; j++ )
            /* Berechnung der Geschwindigkeiten */
	    switch( SURFACE_FLAGS )
	    {
	      case C_N:
		V[i][j] = V[i][j-1] - dely/delx * (U[i][j] - U[i-1][j]);
		if( LEER(FLAG[i-1][j+1]) )
		    U[i-1][j+1] = U[i-1][j] - dely/delx * (V[i][j] - V[i-1][j]);
	      Case C_S:
	        V[i][j-1] = V[i][j] + dely/delx * (U[i][j] - U[i-1][j]);
		if( LEER(FLAG[i-1][j-1]) )
		    U[i-1][j-1] = U[i-1][j] + dely/delx * (V[i][j-1] - V[i-1][j-1]);
	      Case C_O:
		U[i][j] = U[i-1][j] - delx/dely * (V[i][j] - V[i][j-1]);
	        if( LEER(FLAG[i+1][j-1] ) )
		    V[i+1][j-1] = V[i][j-1] - delx/dely * (U[i][j] - U[i][j-1]);
	      Case C_W:
	        U[i-1][j] = U[i][j] + delx/dely * (V[i][j] - V[i][j-1]);
	        if( LEER(FLAG[i-1][j-1]) )
	            V[i-1][j-1] = V[i][j-1] + delx/dely * (U[i-1][j] - U[i-1][j-1]);
	      Case C_NO:
	        U[i][j] = U[i-1][j];
	        V[i][j] = V[i][j-1];
		if( LEER(FLAG[i-1][j+1]) )
		    U[i-1][j+1] = U[i-1][j] - dely/delx * (V[i][j] - V[i-1][j]);
		if( LEER(FLAG[i+1][j-1]) )
		    V[i+1][j-1] = V[i][j-1] - delx/dely * (U[i][j] - U[i][j-1]);
		if( LEER(FLAG[i+1][j+1]) )
		{
		    U[i][j+1] = U[i][j];
		    V[i+1][j] = V[i][j];
		}
	      Case C_NW:
		U[i-1][j] = U[i][j];
		V[i][j] = V[i][j-1];
		if( LEER(FLAG[i-1][j-1]) )
		    V[i-1][j-1] = V[i][j-1] + delx/dely * (U[i-1][j] - U[i-1][j-1]);
		if( LEER(FLAG[i-1][j+1]) )
		{
		    U[i-1][j+1] = U[i-1][j];
		    V[i-1][j] = V[i][j];
		}
	      Case C_SO:
		U[i][j] = U[i-1][j];
		V[i][j-1] = V[i][j];
		if( LEER(FLAG[i-1][j-1]) )
		    U[i-1][j-1] = U[i-1][j] + dely/delx * (V[i][j-1] - V[i-1][j-1]);
		if( LEER(FLAG[i+1][j-1]) )
		{
		    V[i+1][j-1] = V[i][j-1];
		    U[i][j-1] = U[i][j];
		}
	      Case C_SW:		   
	        U[i-1][j] = U[i][j];
		V[i][j-1] = V[i][j];
		if( LEER(FLAG[i-1][j-1]) ) 
		{
		    U[i-1][j-1] = U[i-1][j];
		    V[i-1][j-1] = V[i][j-1];
		}
	      Case C_NS:		   
	        V[i][j] = V[i][j] + delt*GY;
	        V[i][j-1] = V[i][j-1] + delt*GY;
		if( LEER(FLAG[i-1][j+1]) )
		    U[i-1][j+1] = U[i-1][j] - dely/delx * (V[i][j] - V[i-1][j]);
		if( LEER(FLAG[i-1][j-1]) )
		    U[i-1][j-1] = U[i-1][j] + dely/delx * (V[i][j-1] - V[i-1][j-1]);
	      Case C_WO:		    
	        U[i][j] = U[i][j] + delt*GX;
		U[i-1][j] = U[i-1][j] + delt*GX;
		if( LEER(FLAG[i-1][j-1]) )
		    V[i-1][j-1] = V[i][j-1] + delx/dely * (U[i-1][j] - U[i-1][j-1]);
		if( LEER(FLAG[i+1][j-1]) )
		    V[i+1][j-1] = V[i][j-1] - delx/dely * (U[i][j] - U[i][j-1]);
	      Case C_NSWO:		   
	        U[i][j] = U[i][j] + delt*GX;
	        U[i-1][j] = U[i-1][j] + delt*GX;
		V[i][j] = V[i][j] + dely*GY;
		V[i][j-1] = V[i][j-1] + delt*GY;
		if( LEER(FLAG[i+1][j+1]) )
		{
		    U[i][j+1] = U[i][j];
		    V[i+1][j] = V[i][j];
		}
		if( LEER(FLAG[i+1][j-1]) )
		{
		    V[i+1][j-1] = V[i][j-1];
		    U[i][j-1] = U[i][j];
		}
		if( LEER(FLAG[i-1][j-1]) )
		{
		    U[i-1][j-1] = U[i-1][j];
		    V[i-1][j-1] = V[i][j-1];
		}
		if( LEER(FLAG[i-1][j+1]) )
		{
		    V[i-1][j] = V[i][j];
		    U[i-1][j+1] = U[i-1][j];
		}
	      Case C_NSW:
	        V[i][j] = V[i][j] + delt*GY;
	        V[i][j-1] = V[i][j-1] + delt*GY;
		U[i-1][j] = U[i][j] + delx/dely * (V[i][j] - V[i][j-1]);
		if( LEER(FLAG[i-1][j+1]) )
		{
		    U[i-1][j+1] = U[i-1][j];
		    V[i-1][j] = V[i][j];
		}
		if( LEER(FLAG[i-1][j-1]) )
		{
		    U[i-1][j-1] = U[i-1][j];
		    V[i-1][j-1] = V[i][j-1];
	        }
	      Case C_NSO:
	        V[i][j] = V[i][j] + delt*GY;
	        V[i][j-1] = V[i][j-1] + delt*GY;
		U[i][j] = U[i-1][j] - delx/dely * (V[i][j] - V[i][j-1]);
		if( LEER(FLAG[i-1][j+1]) )
		    U[i-1][j+1] = U[i-1][j] - dely/delx * (V[i][j] - V[i-1][j]);
		if( LEER(FLAG[i-1][j-1]) )
		    U[i-1][j-1] = U[i-1][j] + dely/delx * (V[i][j-1] - V[i-1][j-1]);
		if( LEER(FLAG[i+1][j+1]) )
		{
		    U[i][j+1] = U[i][j];
		    V[i+1][j] = V[i][j];
		}
		if( LEER(FLAG[i+1][j-1]) )
		{
		    V[i+1][j-1] = V[i][j-1];
		    U[i][j-1] = U[i][j];
		}
	      Case C_NWO:		  
	        U[i][j] = U[i][j] + delt*GX;
	        U[i-1][j] = U[i-1][j] + delt*GX;
		V[i][j] = V[i][j-1] - dely/delx * (U[i][j] - U[i-1][j]);
		if( LEER(FLAG[i+1][j-1]) )
		    V[i+1][j-1] = V[i][j-1] - delx/dely * (U[i][j] - U[i][j-1]);
		if( LEER(FLAG[i-1][j-1]) )
		    V[i-1][j-1] = V[i][j-1] + delx/dely * (U[i-1][j] - U[i-1][j-1]);
		if( LEER(FLAG[i-1][j+1]) ) 
		{
		    V[i-1][j] = V[i][j];
		    U[i-1][j+1] = U[i-1][j];
		}
		if( LEER(FLAG[i+1][j+1]) )
		{
		    U[i][j+1] = U[i][j];
		    V[i+1][j] = V[i][j];
		}
	      Case C_SWO:		 
	        U[i][j] = U[i][j] + delt*GX;
	        U[i-1][j] = U[i-1][j] + delt*GX;
		V[i][j-1] = V[i][j] + dely/delx * (U[i][j] - U[i-1][j]);
		if( LEER(FLAG[i+1][j-1]) )
		{
		    V[i+1][j-1] = V[i][j-1];
		    U[i][j-1] = U[i][j];
		}
		if( LEER(FLAG[i-1][j-1]) )
		{
		    U[i-1][j-1] = U[i-1][j];
		    V[i-1][j-1] = V[i][j-1];
		}
	      Default:
	        ;/* do nothing */
	    }
      
    for( i = 1; i <= imax; i++ )
	for( j = 1; j <= jmax; j++ )
	    /* Berechnung der Druecke */
	    if( OBERFL(FLAG[i][j]) )
	    switch( SURFACE_FLAGS )
	    {
	      case C_N:
	      case C_S:
		P[i][j] = 2/Re * (V[i][j] - V[i][j-1])/dely;
	      Case C_O:
	      case C_W:
	        P[i][j] = 2/Re * (U[i][j] - U[i-1][j])/delx;
	      Case C_NO:
		P[i][j] = +1/(2*Re) * ( (U[i-1][j] + U[i][j] - U[i][j-1] - U[i-1][j-1]) /dely 
				      + (V[i][j] + V[i][j-1] - V[i-1][j] - V[i-1][j-1]) /delx 
		                      );
	      Case C_NW:
		P[i][j] = -1/(2*Re) * ( (U[i][j] + U[i-1][j] - U[i][j-1] - U[i-1][j-1]) /dely
				      + (V[i+1][j] + V[i+1][j-1] - V[i][j] - V[i][j-1]) /delx
		                      );
		    
	      Case C_SO:		  
		P[i][j] = -1/(2*Re) * ( (U[i][j+1] + U[i-1][j+1] - U[i][j] - U[i-1][j]) /dely
				      + (V[i][j] + V[i][j-1] - V[i-1][j] - V[i-1][j-1]) /delx
		                      );
	      Case C_SW:		    
		P[i][j] = +1/(2*Re) * ( (U[i][j+1] + U[i-1][j+1] - U[i][j] - U[i-1][j]) /dely
				      + (V[i+1][j] + V[i+1][j-1] - V[i][j] - V[i][j-1]) /delx
		                      );
	      Case C_NS:
	      case C_WO:
	      case C_NSWO:
	      case C_NSW:
	      case C_NSO:
	      case C_NWO:
	      case C_SWO:
		P[i][j] = 0.0;
	      Default:
	        ;/* do nothing */
	    }
}

void WRITE_PARTICLES2
(
    char *outputfile,
    int N,
    struct particleline *Particlelines,
    REAL xlength, REAL ylength
)
{
    static int header_written = 0;
    assert( 0 );
    if( ! header_written )
    {
	FILE *f = fopen( outputfile, "w" );
	assert( f );
	fprintf
	( 
	    f, "%f %f\n", 
	    (float) xlength, 
	    (float) ylength
	);
	fclose( f );
	header_written = -1;
    }
    WRITE_PARTICLES( outputfile, N, Particlelines );
}

