/* uvp.c */

#include "flag.h"
#include "uvp.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>

#include "model_compat.h"

#define SQR(x) ((x)*(x))
#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((x)<(y)?(y):(x))
#define abs(x) ((x)<0?-(x):(x))
#define Case break;case
#define Default break;default

#ifdef NDEBUG
#define CHECK_U 
#define CHECK_V 
#define CHECK_P
#define CHECK_RHS 
#define CHECK_F 
#define CHECK_G 
#define CHECK(d)
#else
static void check_array( REAL **a, int imax,  int jmax )
{
    int i, j;
    for( i = 0; i <= imax + 1; i++ )
        for( j = 0; j <= jmax + 1; j++ )
            assert( !( a[i][j] != a[i][j] ) );
}
#define CHECK_U check_array( U,  imax,  jmax )
#define CHECK_V check_array( V,  imax,  jmax )
#define CHECK_P check_array( P,  imax,  jmax )
#define CHECK_RHS check_array( RHS,  imax,  jmax )
#define CHECK_F check_array( F,  imax,  jmax )
#define CHECK_G check_array( G,  imax,  jmax )
#define CHECK(d) (assert( !( (d) != (d) )))
#endif

/**************************************************************************************
 * 
 *     2            u  + u      2      u     + u   2
 *  d(u )      1     ij   i+1,j         i-1,j   ij
 * [-----]  = --( (------------)   - (------------)   ) + 
 *   dx   ij  dx        2                  2
 * 
 * 
 *                 |u  + u     | |u  - u     |     |u     + u  | |u     - u  |
 *              1  | ij   i+1,j| | ij   i+1,j|     | i-1,j   ij| | i-1,j   ij|
 *       alpha -- (------------- -------------  -  ------------- -------------)   , 
 *             dx        2             2                 2             2
 * 
 * 
 *                 v  + v      u  + u          v     + v        u     + u
 *  d(uv)      1    ij   i+1,j  ij   i,j+1      i,j-1   i+1,j-1  i,j-1   ij
 * [-----]  = --( ------------ -----------  -  ---------------- ----------- ) + 
 *   dy   ij  dy       2            2                   2             2
 * 
 * 
 *                 |v  + v     | |u  - u     |     |v     + v       | |u     - u  |
 *              1  | ij   i+1,j| | ij   i,j+1|     | i,j-1   i+1,j-1| | i,j-1   ij|
 *       alpha -- (------------- -------------  -  ------------------ -------------)   , 
 *             dy        2             2                    2               2
 * 
 * 
 *   2      u     - 2 u  + u
 *  d u      i+1,j     ij   i-1,j
 * [---]  = ---------------------
 *    2 ij              2  
 *  dx              (dx)
 * 
 *   2      u     - 2 u  + u
 *  d u      i,j+1     ij   i,j-1
 * [---]  = ---------------------
 *    2 ij              2  
 *  dy              (dy)
 * 
 *          p     - p
 *  dp       i+1,j   ij
 * [--]  = -------------
 *  dx ij        dx      
 *             
 *
 *                                                                           (4)
 * 
 ***************************************************************************************/
#define FORMEL_4a formel_4a( i, j, delx, U, alpha )
static REAL formel_4a
(
    int i, int j,
    REAL delx, 
    REAL **U,
    REAL alpha    
)
{
    REAL t1, t2;
    assert( delx != 0 );

    t1 = (
            SQR( 
	       ( U[i][j] + U[i+1][j] ) / 2 
	    ) 
         -  SQR( 
	       ( U[i-1][j] + U[i][j] ) / 2 
	    )
	 ) 
       / delx;
    
    t2 = (
            abs( U[i][j] + U[i+1][j] )
	    * ( U[i][j] - U[i+1][j] )
	    / 4 
          - 
	    abs( U[i-1][j] + U[i][j] )
	    * ( U[i-1][j] - U[i][j] )
	    / 4
          ) 
	/ delx;
    
    return t1 + alpha * t2;
}
#define FORMEL_4b formel_4b( i, j, dely, U, V, alpha )
static REAL formel_4b
(
    int i, int j,
    REAL dely, 
    REAL **U, REAL **V,
    REAL alpha    
)
{
    REAL t1, t2;
    assert( dely != 0 );
    
    t1 = (
             ( V[i][j] + V[i+1][j] )
	    *( U[i][j] + U[i][j+1] )
	    / 4 
         - 
	     ( V[i][j-1] + V[i+1][j-1] )
	    *( U[i][j-1] + U[i][j] )
	    / 4
	 ) 
       / dely;
    
    t2 = (
             abs( V[i][j] + V[i+1][j] )
	     * ( U[i][j] - U[i][j+1] )
	     / 4 
	 - 
	     abs( V[i][j-1] + V[i+1][j-1] )
	     * ( U[i][j-1] - U[i][j] )
	     / 4
	 ) 
       / dely;
    
    return t1 + alpha * t2;
}
#define FORMEL_4c formel_4c( i, j, delx, U ) 
static REAL formel_4c
(
    int i, int j, 
    REAL delx, REAL **U
)
{
    assert( delx != 0 );
    return
    (
          U[i+1][j]
    -  2* U[i][j]
    +     U[i-1][j]
    )
    / SQR( delx );
}
#define FORMEL_4d formel_4d( i, j, dely, U )
static REAL formel_4d
(
    int i, int j, 
    REAL dely, REAL **U
)
{    
    assert( dely != 0 );
    return
    (
          U[i][j+1]
    -  2* U[i][j]
    +     U[i][j-1]
    )
    / SQR( dely );
}
#define FORMEL_4e formel_4e( i, j, delx, P )
static REAL formel_4e
(
    int i, int j, 
    REAL delx, REAL **P
)
{    
    assert( delx != 0 );
    return ( P[i+1][j] - P[i][j] ) / delx;
}


/**********************************************************************************
 * 
 *                 u  + u      v  + v          u     + u        v     + v
 *  d(uv)      1    ij   i,j+1  ij   i+1,j      i-1,j   i-1,j+1  i-1,j   ij
 * [-----]  = --( ------------ -----------  -  ---------------- ----------- ) + 
 *   dx   ij  dx       2            2                   2             2
 * 
 * 
 *                 |u  + u     | |v  - v     |     |u     + u       | |v     - v  |
 *              1  | ij   i,j+1| | ij   i+1,j|     | i-1,j   i-1,j+1| | i-1,j   ij|
 *       alpha -- (------------- -------------  -  ------------------ -------------)   , 
 *             dx        2             2                    2               2
 * 
 *     2            v  + v      2      v     + v   2
 *  d(v )      1     ij   i,j+1         i,j-1   ij
 * [-----]  = --( (------------)   - (------------)   ) + 
 *   dy   ij  dy        2                  2
 * 
 * 
 *                 |v  + v     | |v  - v     |     |v     + v  | |v     - v  |
 *              1  | ij   i,j+1| | ij   i,j+1|     | i,j-1   ij| | i,j-1   ij|
 *       alpha -- (------------- -------------  -  ------------- -------------)   , 
 *             dy        2             2                 2             2
 * 
 * 
 *   2      v     - 2 v  + v
 *  d v      i+1,j     ij   i-1,j
 * [---]  = ---------------------
 *    2 ij              2  
 *  dx              (dx)
 * 
 *   2      v     - 2 v  + v
 *  d v      i,j+1     ij   i,j-1
 * [---]  = ---------------------
 *    2 ij              2  
 *  dy              (dy)
 * 
 *          p     - p
 *  dp       i,j+1   ij
 * [--]  = -------------
 *  dy ij        dy      
 *             
 * 
 *                                                                           (5)
 * 
 *************************************************************************************/
#define FORMEL_5a formel_5a( i, j, delx, U, V, alpha )
static REAL formel_5a
(
    int i, int j,
    REAL delx,
    REAL **U, REAL **V,
    REAL alpha    
)
{
    REAL t1, t2;
    assert( delx != 0 );
    
    t1 = (
             ( U[i][j] + U[i][j+1] )
	    *( V[i][j] + V[i+1][j] )
	    / 4 
         - 
	     ( U[i-1][j] + U[i-1][j+1] )
	    *( V[i-1][j] + V[i][j] )
	    / 4
	 ) 
       / delx;
    
    t2 = (
             abs( U[i][j] + U[i][j+1] )
	     *( V[i][j] - V[i+1][j] )
	     / 4 
         - 
             abs( U[i-1][j] + U[i-1][j+1] )
	     *( V[i-1][j] - V[i][j] )
	     / 4
         ) 
       / delx;
    
    return t1 + alpha * t2;
}
#define FORMEL_5b formel_5b( i, j, dely, V, alpha )
static REAL formel_5b
(
    int i, int j,
    REAL dely, 
    REAL **V,
    REAL alpha    
)
{
    REAL t1, t2;
    assert( dely != 0 );
    
    t1 = (
            SQR( 
	       ( V[i][j] + V[i][j+1] ) / 2 
	    ) 
         -  SQR( 
	       ( V[i][j-1] + V[i][j] ) / 2 
	    )
	 ) 
       / dely;
    
    t2 = (
            abs( V[i][j] + V[i][j+1] )
	    * ( V[i][j] - V[i][j+1] )
	    / 4 
          - 
	    abs( V[i][j-1] + V[i][j] )
	    * ( V[i][j-1] - V[i][j] )
	    / 4
          ) 
	/ dely;
    
    return t1 + alpha * t2;
}
#define FORMEL_5c formel_5c( i, j, delx, V ) 
static REAL formel_5c
(
    int i, int j, 
    REAL delx, REAL **V
)
{
    assert( delx != 0 );
    return
    (
          V[i+1][j]
    -  2* V[i][j]
    +     V[i-1][j]
    )
    / SQR( delx );
}
#define FORMEL_5d formel_5d( i, j, dely, V )
static REAL formel_5d
(
    int i, int j, 
    REAL dely, REAL **V
)
{    
    assert( dely != 0 );
    return
    (
          V[i][j+1]
    -  2* V[i][j]
    +     V[i][j-1]
    )
    / SQR( dely );
}
#define FORMEL_5e formel_5e( i, j, dely, P )
static REAL formel_5e
(
    int i, int j, 
    REAL dely, REAL **P
)
{    
    assert( dely != 0 );
    return ( P[i][j+1] - P[i][j] ) / dely;
}

/******************************************************
 * 
 *  (n+1)   (n)  dt   (n+1)    (n+1)
 * u     = F   - --( p      - p     )
 *  ij      ij   dx   i+1,j    ij
 *                                 
 *                      i = 1,...,imax-1,  j = 1,...,jmax
 * 
 *                                          (7)
 * 
 ********************************************************/
#define FORMEL_7 formel_7( P, F, imax, jmax, delx, delt, U, FLAG )
static void formel_7
(
    REAL **P, 
    REAL **F, 
    int imax, int jmax, 
    REAL delx, REAL delt, 
    REAL **U, 
    int **FLAG
)
{
    int i, j;
    assert( delx != 0 );
    for( i = 1; i < imax; i++ )
        for( j = 1; j <= jmax; j++ )
	    if( FLUIDX(FLAG[i][j]) && FLUIDX(FLAG[i+1][j]) ) 
		U[i][j] = F[i][j] - delt/delx * ( P[i+1][j] - P[i][j] );
    CHECK_U;
}

/******************************************************
 * 
 *  (n+1)   (n)  dt   (n+1)    (n+1)
 * v     = G   - --( p      - p     )
 *  ij      ij   dy   i,j+1    ij
 *                                 
 *                      i = 1,...,imax,  j = 1,...,jmax-1
 * 
 *                                          (8)
 * 
 ********************************************************/
#define FORMEL_8 formel_8( P, G, imax, jmax, dely, delt, V, FLAG)
static void formel_8
(
    REAL **P, 
    REAL **G, 
    int imax, int jmax, 
    REAL dely, REAL delt, 
    REAL **V, 
    int **FLAG
)
{
    int i, j;
    assert( dely != 0 );
    for( i = 1; i <= imax; i++ )
        for( j = 1; j < jmax; j++ )
	    if( FLUIDX(FLAG[i][j]) && FLUIDX(FLAG[i][j+1]) )
		V[i][j] = G[i][j] - delt/dely * ( P[i][j+1] - P[i][j] );
    CHECK_V;
}

/**********************************************************************
 * 
 * i = 1, ..., imax
 *    j = 1, ..., jmax
 *    
 *    it+1           it
 *   p    = (1-omg) p   +
 *    ij             ij
 *                                it      it+1       it      it+1
 *                               p     + p          p     + p
 *                  omg           i+1,j   i-1,j      i,j+1   i,j-1
 *          ------------------ ( --------------  +  --------------  -  rhs   )
 *               1       1                2                  2            ij
 *           2(----- + -----)         (dx)               (dy)
 *                 2       2     
 *             (dx)    (dy)
 * 
 *                                                            (12)
 * 
 **********************************************************************/
#define FORMEL_12 formel_12( P, imax, jmax, omg, delx, dely, RHS, FLAG )
static void formel_12
(
    REAL **P, 
    int imax, int jmax,
    REAL omg, 
    REAL delx, REAL dely, 
    REAL **RHS, 
    int **FLAG
)
{
    int i, j;
    assert( delx != 0 );
    assert( dely != 0 );
    for( i = 1; i <= imax; i++ )
        for( j = 1; j <= jmax; j++ )
	    if( FLUID(FLAG[i][j]) )
		P[i][j] = ( 1 - omg ) * P[i][j] + 
		 
		    (omg/2 /( 1/SQR(delx) + 1/SQR(dely) ) )
		   *
		    (
			( P[i+1][j] + P[i-1][j] ) / SQR(delx)
		    +   ( P[i][j+1] + P[i][j-1] ) / SQR(dely)
		    -   RHS[i][j]
		    );  
    CHECK_P;	     
}
 
 /*************************************************************************
  * 
  * 
  *         imax jmax
  *         ---  ---   p     - 2 p  + p            p     - 2 p  + p 
  *         \    \      i+1,j     ij   i-1,j        i,j+1     ij   i,j-1            2                1/2
  * res = (  >    >  (-----------------------  +  -----------------------  -  rhs  )  / (imax jmax) )
  *         /    /                  2                          2                 ij
  *         ---  ---            (dx)                       (dy)
  *         i=1  j=1
  *                                                                 (13)
  * 
  ************************************************************************/
#define FORMEL_13 formel_13( imax, jmax, P, delx, dely, RHS, FLAG )
static REAL formel_13
(
    int imax, int jmax,
    REAL **P,  
    REAL delx, REAL dely, 
    REAL **RHS, 
    int **FLAG
)
{
   REAL acc = 0.0;
   int i, j, n = 0;
   assert( delx != 0 );
   assert( dely != 0 );
   assert( imax * jmax != 0 );
   for( i = 1; i <= imax; i++ )
       for( j = 1; j <= jmax; j++ )
           if( FLUID(FLAG[i][j]) )
	   {
	       acc += SQR
	       (
		   ( P[i+1][j] - 2*P[i][j] + P[i-1][j] ) / SQR(delx)
	       +   ( P[i][j+1] - 2*P[i][j] + P[i][j-1] ) / SQR(dely)
	       -   RHS[i][j]
	       );
	       n++;
	   }
    assert( n > 0 );
    return sqrt( acc / n );
}

/**********************************************************************
 * 
 * p  = p  , p        = p      ,    j = 1,...,jmax
 *  0j   1j   imax+1,j   imax,j
 * 
 * p  = p  , p        = p      ,    i = 1,...,imax
 *  i0   i1   i,jmax+1   i,jmax                                (16)
 * 
 **********************************************************************/
#define FORMEL_16 formel_16( P, imax, jmax, FLAG)
static void formel_16
(
    REAL **P, 
    int imax, int jmax, 
    int **FLAG
)
{
    int i, j;
    for( j = 1; j <= jmax; j++ )
    {
	P[0][j] = P[1][j];
	P[imax+1][j] = P[imax][j];
    }
    for( i = 1; i <= imax; i++ )
    {
	P[i][0] = P[i][1];
	P[i][jmax+1] = P[i][jmax];
    }
    /* Berechne zusaetzlich die Randwerte fuer den Druck */
    /* an den innneren Randzellen */
    for( i = 1; i <= imax; i++ )
        for( j = 1; j <= jmax; j++ )
	    switch( BOUNDARY_FLAGS )
	    {
	      case B_N:
	        P[i][j] = P[i][j+1];
	      Case B_W:
	        P[i][j] = P[i-1][j];
	      Case B_S:
	        P[i][j] = P[i][j-1];
	      Case B_O:
	        P[i][j] = P[i+1][j];
	      Case B_NO:
	        P[i][j] = (P[i][j+1] + P[i+1][j]) / 2.0;
	      Case B_SO:
	        P[i][j] = (P[i][j-1] + P[i+1][j]) / 2.0;
	      Case B_NW:
	        P[i][j] = (P[i][j+1] + P[i-1][j]) / 2.0;
	      Case B_SW:
	        P[i][j] = (P[i][j-1] + P[i-1][j]) / 2.0;
	      Default:
	        assert( !VERBOT(FLAG[i][j]) );
	    }
}

/*********************************************************************
 * 
 * F    = u   ,  F      = u         j = 1,...,jmax
 *  0,j    0,j    imax,j   imax,j
 * 
 * G    = v   ,  G      = v         i = 1,...,imax                (17)
 *  i,0    i,0    i,jmax   i, jmax
 *********************************************************************/
#define FORMEL_17 formel_17( U,  V,  F,  G, imax, jmax )
static void formel_17
(
    REAL **U, REAL **V,
    REAL **F, REAL **G, 
    int imax, int jmax
)
{
    int i, j;
    for( j = 1; j <= jmax; j++ )
    {
	F[0][j]    = U[0][j];
	F[imax][j] = U[imax][j];
    }
    for( i = 1; i <= imax; i++ )
    {
	G[i][0]    = V[i][0];
	G[i][jmax] = V[i][jmax];
    }
}

/**********************************************************************
 * 
 * Berechnung von F und G gemaess (9) und (10).
 * Dabei muessen am Rand die Formeln (17) an-
 * gewendet werden.
 * 
 *                      2        2            2
 *                  1  d u      d u        d(u )       d(uv)
 * F   = u   delt( --([---]  + [---]  ) - [-----]   - [-----]   + GX )
 *  ij    ij       Re    2 ij     2 ij      dx   ij     dy   ij
 *                     dx       dy
 * 
 *         i = 1,...,imax,   j = 1,...,jmax-1                      (9)
 * 
 * 
 *                      2        2                        2
 *                  1  d v      d v        d(uv)       d(v )
 * G   = v   delt( --([---]  + [---]  ) - [-----]   - [-----]   + GY )
 *  ij    ij       Re    2 ij     2 ij      dx   ij     dy   ij
 *                     dx       dy
 * 
 *         i = 1,...,imax-1,   j = 1,...,jmax                     (10)
 * 
 * 
 * 
 * 
 * 
 ***********************************************************************/
void COMP_FG
( 
    REAL **U, REAL **V,
    REAL **F, REAL **G, 
    int imax, int jmax,
    REAL delt, REAL delx, REAL dely, 
    REAL GX, REAL GY,
    REAL alpha, REAL Re, 
    int **FLAG
)
{
    int i, j;
    assert( Re != 0 );
    FORMEL_17;
    
    for( i = 1; i < imax; i++ )
        for( j = 1; j <= jmax; j++ )
	    switch( BOUNDARY_FLAGS )
	    {
	      case C_B:
	      case B_N:
	      case B_S:
	        /* do nothing */
	      Case B_W:
	      case B_NW:
	      case B_SW:
	        F[i-1][j] = U[i-1][j];
	      Case B_O:
	      case B_NO:
	      case B_SO:
	        F[i][j] = U[i][j];
	      Default:
	        assert( !VERBOT(FLAG[i][j]) ); 
		if( !LEER(FLAG[i][j]) && !LEER(FLAG[i+1][j]) )
	            F[i][j] = U[i][j] + delt *
	                    (
		                ( FORMEL_4c + FORMEL_4d ) / Re 
		            -   FORMEL_4a
		            -   FORMEL_4b
		            +   GX
		            );
	    }
    CHECK_F;
    for( i = 1; i <= imax; i++ )
        for( j = 1; j < jmax; j++ )
	    switch( BOUNDARY_FLAGS )
	    {
	      case C_B:
	      case B_W:
	      case B_O:
	        /* do nothing */
	      Case B_N:
	      case B_NW:
	      case B_NO:
	        G[i][j] = V[i][j];
	      Case B_S:
	      case B_SW:
	      case B_SO:
	        G[i][j-1] = V[i][j-1];
	      Default:
	        assert( !VERBOT(FLAG[i][j]) ); 
		if( !LEER(FLAG[i][j]) && !LEER(FLAG[i][j+1]) )
		    G[i][j] = V[i][j] + delt *
			    (
			        ( FORMEL_5c + FORMEL_5d ) / Re
			    -   FORMEL_5a
			    -   FORMEL_5b
			    +   GY
			    );
	    }
   CHECK_G;
}


/**********************************************************************
 * 
 * Berechnung der rechten Seite der Druckgleichung (11)
 * 
 * 
 *    (n+1)     (n+1)   (n+1)        (n+1)     (n+1)   (n+1)
 *   p     - 2 p     + p            p     - 2 p     + p
 *    i+1,j     ij      i-1,j        i,j+1     ij      i,j-1
 *  --------------------------  +  --------------------------  =
 *                2                              2
 *            (dx)                           (dy)
 * 
 * 
 *                              (n)   (n)          (n)   (n)
 *                             F   - F            G   - G 
 *                         1    ij    i-1,j        ij    i,j-1
 *                        --( --------------  +  -------------- )
 *                        dt        dx                  dy
 * 
 *                                  i = 1,...,imax,  j = 1,...,jmax
 * 
 * 
 *                                                             (11)
 *                                                          
 * 
 ************************************************************************/
void COMP_RHS
(
    REAL **F, REAL **G, 
    REAL **RHS, 
    int imax, int jmax, 
    REAL delt, REAL delx, REAL dely
)
{
   int i, j;
   assert( delx != 0 );
   assert( dely != 0 );
   assert( delt != 0 );
   for( i = 1; i <= imax; i++ )
       for( j = 1; j <= jmax; j++ )
           RHS[i][j] = 
            (
               ( F[i][j] - F[i-1][j] ) / delx 
            +
               ( G[i][j] - G[i][j-1] ) / dely
            ) / delt;
    CHECK_RHS;	     
}
 
 /**************************************************************************************
  * 
  * SOR-Iteration fuer die Druck-(Poisson-)Gleichung gemaess (12). In jeder 
  * Iteration muessen zumaechst die Randwerte fuer P gemaess (16) gesetzt werden.
  * Die Iteration wird abgebrochen, wenn die Norm des Residuums res die Toleranzgrenze
  * eps unterschreitet oder die maximale Iterationsanzahl itermax erreicht ist. Nach
  * dem Abbruch soll die Anzahl der benoetigten Iterationen zurueckgegeben und in res
  * die aktuelle Groesse des Residuums stehen.
  *
  **************************************************************************************/
int POISSON
(
    REAL **P, REAL **RHS, 
    REAL **U, REAL **V, 
    int imax, int jmax,
    REAL delx, REAL dely,
    REAL eps, int itermax,
    REAL omg, REAL *res, 
    int **FLAG
)
{
    /* Setze */ int it = 0; *res = 2.0 * eps;
    set_state( CS_SOR );
    /* Solange */ while( it < itermax && *res > eps )
    {
	set_progres( (float)it / (float)itermax );
	/* Setze die Randwerte fuer den Druck gemaess (16) */
	FORMEL_16;
	/* Fuehre einen SOR-Zyklus gemaess (12) durch */
	FORMEL_12;
	/* Berechne das Residuum res der Druckgleichung gemaess (13) */
	*res = FORMEL_13;
	assert( !(*res != *res) );
	it = it + 1;
    }
    return it;
}

/*************************************************************
 * 
 * Die neuen Geschwindigkeiten werden gemaess (7) und (8)
 * berechnet.
 * 
 **************************************************************/
 
void ADAP_UV
(
    REAL **U, REAL **V, 
    REAL **F, REAL **G, 
    REAL **P, 
    int imax, int jmax, 
    REAL delt, 
    REAL delx, REAL dely, 
    int **FLAG
)
{
    FORMEL_7;
    FORMEL_8;
}
