/* visual.c */

#include "flag.h"
#include "visual.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "model_compat.h"

static char Outputfile[] = "output.bin";

static void write_float( FILE *f,  REAL r )
{
    float t = r;
    int success = fwrite( &t, sizeof(float), 1, f );
    assert( success );
    /* fprintf( f, "\n" ); */
}
/***************************************************************
 * 
 * Ausgabe der Daten zur Visualisierung.
 * 
 ***************************************************************/
void OUTPUTVEC
(
    REAL **U, REAL **V, REAL **P, 
    REAL xlength, REAL ylength, 
    int imax, int jmax, 
    int **FLAG, 
    REAL **PSI, REAL **ZETA
)
{
    FILE *f;
    int i, j;
    assert( 0 );
    f = fopen( Outputfile, "w" );
    assert( f );
    fprintf( f, "%f\n%f\n%d\n%d\n", 
        (float)xlength, (float)ylength, imax, jmax );
    for( j = 1; j <= jmax; j++ )
        for( i = 1; i <= imax; i++ )
	    if( FLUID(FLAG[i][j]) )
	        write_float( f, ( U[i-1][j] + U[i][j] ) / 2 );
	    else write_float( f, 0.0 );
    for( j = 1; j <= jmax; j++ )
        for( i = 1; i <= imax; i++ )
	    if( FLUID(FLAG[i][j]) )
		write_float( f, ( V[i][j-1] + V[i][j] ) / 2 );
	    else write_float( f, 0.0 );
    for( j = 1; j <= jmax; j++ )
        for( i = 1; i <= imax; i++ )
	    if( FLUID(FLAG[i][j])  )
		write_float( f, P[i][j] );
	    else write_float( f, 0.0 );
    for( j = 1; j <= jmax; j++ )
        for( i = 1; i <= imax; i++ )
	    write_float( f, PSI[i][j] );
    for( j = 1; j <= jmax; j++ )
        for( i = 1; i <= imax; i++ )
	    write_float( f, ZETA[i][j] );
    fclose( f );
}

struct particleline *SET_PARTICLES
(
    int N,
    REAL x1, REAL y1, REAL x2, REAL y2
)
{
    int i;
    struct particleline *result;
    return 0;
    result = (struct particleline *) malloc( N * sizeof( struct particleline ) );
    assert( result );
    for( i = 0; i < N; i++ )
    {
        result[i].length = 1;
	result[i].Particles = (struct particle *) 
	    malloc( sizeof( struct particle ) );
	assert( result[i].Particles );
	result[i].Particles->x = ((float)i / (float)N) * (x2 - x1) + x1;
	result[i].Particles->y = ((float)i / (float)N) * (y2 - y1) + y1;
	result[i].Particles->next = (struct particle *) 
	    malloc( sizeof( struct particle ) );
	assert( result[i].Particles->next );
	result[i].Particles->next->x = ((float)i / (float)N) * (x2 - x1) + x1;
	result[i].Particles->next->y = ((float)i / (float)N) * (y2 - y1) + y1;
	result[i].Particles->next->next = 0;
	assert( result[i].Particles && result[i].Particles->next );
     }
     return result;
}

void ADVANCE_PARTICLES
(
    int imax, int jmax, 
    REAL delx, REAL dely, REAL delt,
    REAL **U, REAL **V,
    int N, struct particleline *Particlelines, 
    int **FLAG
)
{
    int n;
    int sum = 0;
    struct Particle *p, **lastP;
    struct ParticleLine *pl;
    int state;
#ifdef INTEREST
    printf( "  Advancing particles " );
#endif
    for( state = -1; state <= 0; state++ )
    {
    set_state( state ? CS_FLUIDS : CS_PARTICLES );
    for( n = 0; n < (state ? FLUID_FLAVORS : PARTICLE_FLAVORS); n++ )
	for
	(
	    lastP = 
		state ? &model.fluids[n].first : &model.particles[n].first,
	      p = *lastP;
	    p; 
	    p = *lastP
	)
	{
	    REAL x = p->x, y = p->y;
	    REAL u, v;
	    int i, j;
	    REAL x1, y1, x2, y2;
	    REAL u1, u2, u3, u4;
	    REAL v1, v2, v3, v4;
	    int flag;
	    
	    /* Berechnung von u(x,y) */
	    
	    i = (int)(x / delx) + 1;
	    j = (int)(y / dely + 0.5) + 1;
	    
	    x1 = (i-1) * delx;
	    y1 = (j-1.5) * dely;
	    x2 = i * delx;
	    y2 = (j-0.5) * dely;
	    
	    u1 = U[i-1][j-1];
	    u2 = U[i][j-1];
	    u3 = U[i-1][j];
	    u4 = U[i][j];
	    
	    u = (
	            (x2-x)*(y2-y)*u1
		+   (x-x1)*(y2-y)*u2
		+   (x2-x)*(y-y1)*u3
		+   (x-x1)*(y-y1)*u4
	        )
	      / (delx * dely);
	      
	    /* Berechnung von v(x,y) */
	    
	    i = (int)(x / delx + 0.5) + 1;
	    j = (int)(y / dely) + 1;
	    
	    x1 = (i-1.5) * delx;
	    y1 = (j-1) * dely;
	    x2 = (i-0.5) * delx;
	    y2 = j * dely;

	    v1 = V[i-1][j-1];
	    v2 = V[i][j-1];
	    v3 = V[i-1][j];
	    v4 = V[i][j];
	    
	    v = (
	            (x2-x)*(y2-y)*v1
		+   (x-x1)*(y2-y)*v2
		+   (x2-x)*(y-y1)*v3
		+   (x-x1)*(y-y1)*v4
	        )
	      / (delx * dely);
	      
	    /* Bestimme die neue Position des Partikels */
	    
	    p->x = x + delt * u;
	    p->y = y + delt * v;
	    
	    /* Verlassen die Partikel das Grundgebiet, oder laufen */
	    /* laufen in Hinderniszellen, so werden sie aus der */
	    /* Liste entfernt, und ihr Speicher wird freigeben. */
	    
	    x = p->x; y = p->y;
	    
	    flag = 0;
	    flag = flag || ( x < 0.0 || x > delx * imax );
	    flag = flag || ( y < 0.0 || y > dely * jmax );
	    if( !flag )
	    {
		i = (int)(x / delx) + 1;
		j = (int)(y / dely) + 1;
	        flag = flag || ( HINDER(FLAG[i][j]) );
	    }
	    if( flag )
	    {
		*lastP = p->next;
		free( p );
#ifdef INTEREST
		printf( "*" );
#endif
	    } else {
	        lastP = & p->next;
#ifdef INTEREST
		printf( "." );
#endif
	    }
	    sum++;
	}}
#ifdef INTEREST
    printf( "(%d)\n", sum );
#endif
}

void INJECT_PARTICLES
(
    int N,
    struct particleline *Particlelines
)
{
    int n;
    int sum = 0;
    struct Particle *q;
#ifdef INTEREST
    printf( "  Injecting particles" );
#endif
    for( n = 0; n < PARTICLE_FLAVORS; n++ )
    for( q = model.particles[n].injection; q; q = q->next )
    {
	struct Particle *p = (struct Particle *) 
	    malloc( sizeof(struct Particle) );
	assert( p );
	*p = *q;
	p->next = model.particles[n].first;
	model.particles[n].first = p;
#ifdef INTEREST
	printf( "." );
#endif
	sum++;
    }
#ifdef INTEREST
    printf( "(%d)\n", sum );
#endif
}

void WRITE_PARTICLES
(
    char *outputfile, 
    int N, 
    struct particleline *Partlines
)
{
    int n, sum = 0;
    struct particle *p;
    FILE *f;
    assert( 0 );
    f = fopen( outputfile, "a" );
    assert( f );
    printf( "  Ausgabe der Partikel." );
    for( n = 0; n < N; n++ )
        sum += Partlines[n].length;
    fprintf( f, "%d\n", sum );
    for( n = 0; n < N; n++ )
        for( p = Partlines[n].Particles->next; p; p = p->next )
	{
	    write_float( f, p->x );
	    write_float( f, p->y ); 
	}   
    fclose( f );
}

void PARTICLE_TRACING
(
    char *outputfile, 
    REAL delt_trace, 
    REAL t, 
    int imax, int jmax, 
    REAL delx, REAL dely, REAL delt, 
    REAL **U, REAL **V, 
    int N, 
    struct particleline *Partlines, 
    REAL xlength, REAL ylength, 
    int **FLAG
)
{
    static REAL last_t = 0.0;
    static int header_written = 0;
    return;
    if( ! header_written )
    {
        FILE *f = fopen( outputfile, "w" );
	assert( f );
	fprintf
	( 
	    f, "%f %f\n", 
	    (float) xlength, 
	    (float) ylength
	);
	fclose( f );
	header_written = -1;
    }
    ADVANCE_PARTICLES
    (
        imax, jmax, 
	delx, dely, delt, 
	U, V, 
	N, Partlines, 
	FLAG
    );
    if( t - last_t > delt_trace )
    {
        WRITE_PARTICLES( outputfile, N, Partlines );
	last_t = t;
    }
}

void STREAKLINES
(
    char *outputfile, 
    REAL delt_inject, 
    REAL delt_streak, 
    REAL t, 
    int imax, int jmax, 
    REAL delx, REAL dely, REAL delt, 
    REAL **U, REAL **V, 
    int N, 
    struct particleline *Partlines, 
    REAL xlength, REAL ylength, 
    int **FLAG      
)
{
    static REAL last_t = 0.0;
    static REAL last_inject = 0.0;
    static int header_written = 0;
    return;
    if( ! header_written )
    {
        FILE *f = fopen( outputfile, "w" );
	assert( f );
	fprintf
	( 
	    f, "%f %f\n", 
	    (float) xlength, 
	    (float) ylength
	);
	fclose( f );
	header_written = -1;
    }
    ADVANCE_PARTICLES
    (
        imax, jmax, 
	delx, dely, delt, 
	U, V, 
	N, Partlines, 
	FLAG
    );
    if( t - last_inject > delt_inject )
    {
	INJECT_PARTICLES( N, Partlines );
	last_inject = t;
    }
    if( t - last_t > delt_streak )
    {
        WRITE_PARTICLES( outputfile, N, Partlines );
	last_t = t;
    }    
}

void COMP_PSI_ZETA
(
    REAL **U, REAL **V, 
    REAL **PSI, REAL **ZETA, 
    int **FLAG,
    int imax, int jmax, 
    REAL delx, REAL dely
)
{
    int i, j;
    printf( "Berechnung von psi und zeta.\n" );
    for( i = 1; i <= imax; i++ )
    {
	PSI[i][0] = 0.0;    PSI[i][jmax+1] = 0.0;
	ZETA[i][0] = 0.0;   ZETA[i][jmax+1] = 0.0;
    }
    for( j = 1; j <= jmax; j++ )
    {
	PSI[0][j] = 0.0;    PSI[imax+1][j] = 0.0;
	ZETA[0][j] = 0.0;   ZETA[imax+1][j] = 0.0;
    }
    for( i = 1; i <= imax; i++ )
        for( j = 1; j <= jmax; j++ )
	    if( FLUID( FLAG[i][j] ) )
	    {
		PSI[i][j] = PSI[i][j-1] + U[i][j] * dely;
		ZETA[i][j] = (U[i][j+1] - U[i][j]) / dely 
		           - (V[i+1][j] - V[i][j]) / delx;
	    }
	    else if( HINDER( FLAG[i][j] ) )
	    {
		PSI[i][j] = PSI[i][j-1];
	        ZETA[i][j] = 0.0;
	    }
}
