/* Slider.c */

#include "../glib.h"

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/ScrollBar.h>
#include <Xm/TextF.h>

#include "Slider.h"

typedef struct {
    Widget form;
    Widget scrollBar, textField;
    int ratio;
    float min, max;
    float value;
    XtCallbackProc call_proc;
    XtPointer client_data;
    int flag;
} SliderRec;
typedef SliderRec *SliderRecPtr;

typedef struct {
    int reason;
    XEvent *event;
    float value;
} SliderCallbackStruct;

static int ififty = 75;
static float fzero = 0.0;
static float fhundred = 110.0;
static float ffifty = 55.0;

static XtResource item_resources[] = 
{
    { XmNwidthRatio, XmCWidthRatio, XmRInt, sizeof(int),
      XtOffset( SliderRecPtr, ratio ), XmRInt, (XtPointer) &ififty },
    { XmNmin, XmCMin, XmRFloat, sizeof(float),
      XtOffset( SliderRecPtr, min ), XmRFloat, (XtPointer) &fzero },
    { XmNmax, XmCMax, XmRFloat, sizeof(float),
      XtOffset( SliderRecPtr, max ), XmRFloat, (XtPointer) &fhundred },
    { XmNvalue, XmCValue, XmRFloat, sizeof(float),
      XtOffset( SliderRecPtr, value ), XmRFloat, (XtPointer) &ffifty },
};

static void CallCallback
(
    SliderRecPtr rec,
    XEvent *event
)
{
    SliderCallbackStruct cb;

    if( rec->call_proc != NULL )
    {
        cb.reason = XmCR_VALUE_CHANGED;
	cb.event = event;
	cb.value = rec->value;
	(*rec->call_proc)( rec->form, rec->client_data, &cb );
    }
}

static void ScrollBarCallback
(
    Widget widget,
    XtPointer client_data,
    XtPointer call_data
)
{
    XmScrollBarCallbackStruct *cb = 
      (XmScrollBarCallbackStruct *) call_data;
    SliderRecPtr rec = (SliderRecPtr) client_data;
    float val = (float) cb->value / 100.0;
    char buffer[50];
    
    if( rec->flag )
    {
        int save_flag = rec->flag;
        rec->flag = FALSE;

        rec->value = rec->min * (1.0 - val) + rec->max * val;

	sprintf( buffer, "%g", rec->value );
	XmTextFieldSetString( rec->textField, buffer );

	rec->flag = save_flag;

	CallCallback( rec, cb->event );
    }
}

static void TextFieldCallback
(
    Widget widget,
    XtPointer client_data,
    XtPointer call_data
)
{
    XmAnyCallbackStruct *cb = 
      (XmAnyCallbackStruct *) call_data;
    SliderRecPtr rec = (SliderRecPtr) client_data;
    char *contents = XmTextFieldGetString( widget );
    float val;
    int value, size, increment, page;

    if( rec->flag )
    {
        int save_flag = rec->flag;
        rec->flag = FALSE;

	sscanf( contents, "%f", &val );

	XmScrollBarGetValues( rec->scrollBar,
          &value, &size, &increment, &page );
	value = (val - rec->min) / (rec->max - rec->min) * 100;
	XmScrollBarSetValues
	( 
	    rec->scrollBar,
	    value, size, increment, page,
	    FALSE
	);

	XtFree( contents );

	rec->flag = save_flag;

	CallCallback( rec, cb->event );
    }
}

Widget CreateSlider
(
    Widget parent,
    String name,
    ArgList add_args,
    Cardinal num_args,
    XtCallbackProc proc,
    XtPointer client_data
)
{
    SliderRecPtr rec;
    Arg args[10];
    char buffer[100];

    rec = (SliderRecPtr) XtMalloc( sizeof(SliderRec) );
    rec->call_proc = proc;
    rec->client_data = client_data;
    rec->flag = TRUE;

    rec->form = XmCreateForm
    (
        parent,
	name,
	add_args, num_args
    );
    
    XtGetSubresources
    (
        rec->form,
	rec, 
	name,
	SLIDER_CLASS,
	item_resources,
	XtNumber(item_resources),
	add_args, num_args
    );

    strcpy( buffer, name );
    strcat( buffer, "_scrollBar" );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[2], XmNbottomAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_POSITION );
    XtSetArg( args[4], XmNrightPosition, rec->ratio );
    XtSetArg( args[5], XmNorientation, XmHORIZONTAL );
    rec->scrollBar = XmCreateScrollBar
    (
        rec->form,
	buffer,
	args, 6
    );
    XtAddCallback
    (
        rec->scrollBar,
	XmNvalueChangedCallback,
	ScrollBarCallback,
	(XtPointer) rec
    );
    strcpy( buffer, name );
    strcat( buffer, "_textField" );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[2], XmNleftWidget, rec->scrollBar );
    XtSetArg( args[3], XmNbottomAttachment, XmATTACH_FORM );
    XtSetArg( args[4], XmNrightAttachment, XmATTACH_FORM );
    rec->textField = XmCreateTextField
    (
        rec->form,
	buffer,
	args, 5
    );
    XtAddCallback
    (
        rec->textField,
	XmNvalueChangedCallback,
	TextFieldCallback,
	(XtPointer) rec
    );

    SliderSetCurrent( rec->form, rec->value, FALSE );

    XtManageChildren( &rec->scrollBar, 2 );

    return rec->form;
}

static SliderRecPtr GetRec( Widget slider )
{
    WidgetList widgets;
    XtCallbackList callbacks;
    Arg args[5];
    int i;
    
    XtSetArg( args[0], XmNchildren, &widgets );
    XtGetValues( slider, args, 1 );
    XtSetArg( args[0], XmNvalueChangedCallback, &callbacks );
    XtGetValues( widgets[0], args, 1 );
    for( i = 0; ; i++ )
    {
	if( callbacks[i].callback == ScrollBarCallback )
	    return (SliderRecPtr) callbacks[i].closure;
	if( callbacks[i].callback == TextFieldCallback )
	    return (SliderRecPtr) callbacks[i].closure;
	if( callbacks[i].callback == NULL )
	{
	    fprintf( stderr, "Slider.c: something went wrong" );
	    exit( 1 );
	}
    }
}

void SliderSetCurrent( Widget slider, float value, int notify )
{
    SliderRecPtr rec = GetRec( slider );
    int save_flag = rec->flag;
    char buffer[80];
    int v, s, i, p;
    
    rec->flag = FALSE;
    
    rec->value = value;
    
    XmScrollBarGetValues( rec->scrollBar,
        &v, &s, &i, &p );
    v = (value - rec->min) / (rec->max - rec->min) * 100;
    XmScrollBarSetValues
    ( 
	rec->scrollBar,
	v, s, i, p,
	FALSE
    );
    
    sprintf( buffer, "%g", rec->value );
    XmTextFieldSetString( rec->textField, buffer );
    
    rec->flag = save_flag;

    CallCallback( rec, NULL );
}

void SliderGetCurrent( Widget slider, float *value_return )
{
    SliderRecPtr rec = GetRec( slider );
    
    *value_return = rec->value;
}
