/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      callbacks.c                                            *
*                                                                    *
*  Contents:  The implementation of the manipulation specific part   *
*             of the floatator program.                              *
*             This part of the program is concerned with             *
*             maniulating the scene by use of Xmoti.                 *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  01.08.95 First steps                             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

/********************************************************************\
*                                                                    *
*                              Includes                              *
*                                                                    *
\********************************************************************/

#include "../glib.h"

#include "ctrl.h"
#include "view.h"

#include <unistd.h>

#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/MainW.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>

#include "icons/progres.xbm"

/********************************************************************\
*                                                                    *
*                           Hilfsroutinen                            *
*                                                                    *
\********************************************************************/

/********************************************************************\
*                                                                    *
*                             Callbacks                              *
*                                                                    *
\********************************************************************/

void FloatatorCallback
(
    XtPointer closure,
    int *source,
    XtInputId *id
)
{
    int message, tmp, msg;
    int nbytes;
    char buffer[80];
    FILE *f;
        
    nbytes = read( *source, &message, sizeof(message) );
    if( nbytes == sizeof(message) )
    {
	switch( message )
	{
	  case FDCMPTime:
	    nbytes = read( *source, &model.time, sizeof(model.time) );
	    sprintf( buffer, "%-7.7g", (float)model.time ); 
	    xms = XmStringCreateLtoR
	    ( 
		buffer,
		XmFONTLIST_DEFAULT_TAG
	    );
	    XtSetArg( args[0], XmNlabelString, xms );
	    XtSetValues( widget.Main.Statusline.time, args, 1 );
	    XmStringFree( xms ); 
	  Case FDCMPState:
	    nbytes = read( *source, &model.state, sizeof(model.state) );
	    XtSetArg( args[0], XmNlabelString, fetch_state_string() );
	    XtSetArg( args[1], XmNforeground, fetch_state_color() );
	    XtSetValues( widget.Main.Statusline.state, args, 2 );
	  Case FDCMPProgres:
	    nbytes = read( *source, &model.progres, sizeof(model.progres) );
	    ProgresCallback( widget.Main.Statusline.time, NULL, NULL );
	  Case FDCMPResiduum:
	    nbytes = read( *source, &model.res, sizeof(model.res) );
	    sprintf( buffer, "%7g", (float)model.res ); 
#if 0	
	    /* Why the heck does this not work?!? */
	    xms = XmStringCreateLtoR
	    ( 
		buffer,
		XmFONTLIST_DEFAULT_TAG
	    );
	    XtSetArg( args[0], XmNlabelString, xms );
	    XtSetArg( args[1], XmNforeground, fetch_residuum_color() );
	    XtSetValues( widget.Main.Statusline.residuum, args, 2 );
	    XmStringFree( xms ); 
#endif
	  Case FDCMPDataSaved:
	    free( model.output_filename );
	    read( *source, &tmp, sizeof(tmp) );
	    model.output_filename = (char *) malloc( tmp+1 );
	    if( model.output_filename == 0 )
	    {
		fprintf( stderr, "Out of memory" );
		exit( 1 );
	    }
	    read( *source, model.output_filename, tmp );
	    model.output_filename[tmp] = 0;
#if 0
	    sleep( 1 );
#endif	    
	    f = fopen( model.output_filename, "r" );
	    if( f == 0 )
	    {
		fprintf( stderr, "Datei \"%s\" konnte nicht geoeffnet werden.\n",  model.output_filename );
		exit( 1 );
	    }
	    read_DataPackage_from_file( &model, f, TRUE );
	    fclose( f );
	    unlink( model.output_filename );
	    msg = FDCMPDataRead;
	    write( *source, &msg, sizeof(msg) );
	    update_view(); 
	  Default:
	    printf( "message %d (%c) received.\n", message, (char)message );
	}
    }
}

void FileNewCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    int msg = FDCMPNewFileChoices;
    int num_choices, i;
    struct FDCMPNewFileChoices_Params *p;
    XmStringTable st; 
    
    write( floatator_socket, &msg, sizeof(msg) );
    read( floatator_socket, &num_choices, sizeof(num_choices) );
    p = malloc( sizeof(struct FDCMPNewFileChoices_Params) +
	    (num_choices - 1) *
	    FDCMP_NEW_FILE_CHOICE_NAMELENGTH );
    if( p == 0 )
    {
	fprintf( stderr,  "Out of memory" );
	exit( 1 );
    }
    read
    ( 
	floatator_socket, 
	p->choice[0].name, 
	num_choices * FDCMP_NEW_FILE_CHOICE_NAMELENGTH 
    );
    p->num_choices = num_choices;
    st = (XmStringTable) XtMalloc( num_choices * sizeof(XmString) );
    for( i = 0; i < num_choices; i++ )
    {
	xms = XmStringCreateLtoR
	(
	    p->choice[i].name, 
	    XmFONTLIST_DEFAULT_TAG
	);
	if( xms == NULL )
	{
	    fprintf( stderr, "Out of memory" );
	    exit( 1 );
	}
	st[i] = xms;
    }
    XtSetArg( args[0], XmNlistItemCount, num_choices );
    XtSetArg( args[1], XmNlistItems, st );
    XtSetValues( widget.nnew, args, 2 );
    XtFree( (XtPointer) st );
    free( p );
    
    XtManageChild( widget.nnew );
}

void FileOpenCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
printf( "+ FileOpenCallback\n" );
fflush( stdout );
    XtManageChild( widget.open );
}

void FileSaveCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void FileSaveAsCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtManageChild( widget.saveAs );
}

void FileQuitCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    int msg = FDCMPQuit;
    
    if( everything_saved() )
    {
	write( floatator_socket, &msg, sizeof(msg) );
        exit( 0 );
    }
}

void EditCutCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void EditCopyCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void EditPasteCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void EditPropertiesCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtManageChild( widget.properties );
}

void OptionsComputationCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtManageChild( widget.computation );
}

void OptionsGeometryCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void OptionsTimeCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void OptionsPressureCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void OptionsProblemCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void OptionsBoundaryCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void OptionsParticlesCallback
( 
    Widget widget, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void HelpAboutCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtManageChild( widget.about );
}

void UpdateViewCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmDrawingAreaCallbackStruct *cb = 
	(XmDrawingAreaCallbackStruct *)call_data;

    if( call_data == NULL )
    {
	XClearWindow
	( 
	    display, 
	    XtWindow( widget.Main.Statusline.progres ) 
	);
    } else {
	if( cb->event == NULL ) return;
	if( cb->event->xexpose.count != 0 ) return;
    }
    update_view();   
}

void MouseClickCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmDrawingAreaCallbackStruct *cb = 
	(XmDrawingAreaCallbackStruct *)call_data;
    float sx, sy;
    REAL x, y;
    int i, j;
    XWindowAttributes attribs;
    int msg;
    struct Particle *p;

    switch( cb->event->type )
    {
      case ButtonRelease:
	sx = cb->event->xbutton.x;
	sy = cb->event->xbutton.y;
	XGetWindowAttributes
	( 
	    display, 
	    XtWindow( widget.Main.view ), 
	    &attribs 
	);
	x = (sx/attribs.width * view.xwidth + view.xoffset) * model.parms.xlength;
	y = (1.0-(sy/attribs.height * view.yheight + view.yoffset)) * model.parms.ylength;
	i = x/model.parms.xlength * model.imax + 1;
	j = y/model.parms.ylength * model.jmax + 1;
	
	switch( ctrl.mode )
	{
	  case EM_FLUID:
	    msg = FDCMPNewFluid;
	    write( floatator_socket, &msg, sizeof(msg) );
	    write( floatator_socket, &ctrl.flavor, sizeof(ctrl.flavor) );
	    write( floatator_socket, &i, sizeof(i) );
	    write( floatator_socket, &j, sizeof(j) );
	  Case EM_PART:
	    msg = FDCMPNewParticle;
	    write( floatator_socket, &msg, sizeof(msg) );
	    write( floatator_socket, &ctrl.flavor, sizeof(ctrl.flavor) );
	    write( floatator_socket, &x, sizeof(x) );
	    write( floatator_socket, &y, sizeof(y) );
	    p = (struct Particle *)malloc( sizeof(struct Particle) );
	    if( p == 0 )
	    {
	        fprintf( stderr, "Out of memory\n" );
		exit( 1 );
	    }
	    p->x = x;
	    p->y = y;
	    p->next = model.particles[ctrl.flavor].first;
	    model.particles[ctrl.flavor].first = p;
	  Case EM_OBST:
	    msg = FDCMPNewObstacle;
	    write( floatator_socket, &msg, sizeof(msg) );
	    write( floatator_socket, &i, sizeof(i) );
	    write( floatator_socket, &j, sizeof(j) );
	    model.FLAG[i][j] |= B_C;
	  Case EM_CLEAR:
	    msg = FDCMPEmptyArea;
	    write( floatator_socket, &msg, sizeof(msg) );
	    write( floatator_socket, &x, sizeof(x) );
	    write( floatator_socket, &y, sizeof(y) );
	    write( floatator_socket, &i, sizeof(i) );
	    write( floatator_socket, &j, sizeof(j) );
	}
    }
    update_view();
}

void HScrollCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void VScrollCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
}

void ProgresCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    static int initialized = FALSE;
    static GC gc;
    static XGCValues gc_values;
    XWindowAttributes attribs;
    int x, y;
    unsigned int width, height;
    Dimension marginWidth, marginHeight;
    Pixel foreground, background;
    XmDrawingAreaCallbackStruct *cb = 
	(XmDrawingAreaCallbackStruct *)call_data;

    if( call_data == NULL )
    {
	XClearWindow
	( 
	    display, 
	    XtWindow( widget.Main.Statusline.progres ) 
	);
    } else {
	if( cb->event == NULL ) return;
	if( cb->event->xexpose.count != 0 ) return;
    }
    if( ! initialized )
    {
	Pixmap pixmap;
	
	gc = XtGetGC
	(
	    widget.Main.Statusline.progres, 
	    0, 
	    &gc_values
	);
	XtSetArg( args[0], XmNforeground, &foreground );
	XtSetArg( args[1], XmNbackground, &background );
	XtGetValues( widget.Main.Statusline.progres, args, 2 );
	pixmap = XCreatePixmapFromBitmapData
	(
	    display, DefaultRootWindow(display),
	    progres_bits, progres_width, progres_height,
	    foreground, background,
	    DefaultDepthOfScreen
	    ( 
		ScreenOfDisplay
		( 
		    display, 
		    DefaultScreen( display ) 
		) 
	    )
	);
	XSetTile( display, gc, pixmap );
	XSetFillStyle( display, gc, FillTiled );
	
	initialized = TRUE;
    }
    
    XGetWindowAttributes
    ( 
	display, 
	XtWindow( widget.Main.Statusline.progres ), 
	&attribs 
    );
    XtSetArg( args[0], XmNmarginWidth, &marginWidth );

    XtSetArg( args[1], XmNmarginHeight, &marginHeight );
    XtGetValues( widget.Main.Statusline.progres, args, 2 );
    x = marginWidth;
    y = marginHeight;
    width = (attribs.width - 2 * marginWidth) * model.progres;
    height = attribs.height - 2 * marginHeight;
    XSetTSOrigin
    (
	display, 
	gc, 
	x, y
    );
    XFillRectangle
    (
	display, 
	XtWindow( widget.Main.Statusline.progres ), 
	gc, 
	x, y, 
	width, height
    );
}

void NewOKCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmSelectionBoxCallbackStruct *cb = 
	(XmSelectionBoxCallbackStruct *)call_data;
    XmStringTable items;
    int num_items;
    int msg = FDCMPNewFile;
    int choice, i;    

    XtSetArg( args[0], XmNlistItems, &items );
    XtSetArg( args[1], XmNlistItemCount, &num_items );
    XtGetValues( w, args, 2 );
    for( i = 0; i < num_items; i++ )
	if( XmStringCompare( items[i], cb->value ) )
	    break; 
    
    XtUnmanageChild( widget.nnew );
    
    XtSetArg( args[0], XmNlistItemCount, 0 );
    XtSetArg( args[1], XmNlistItems, NULL );
    XtSetValues( widget.nnew, args, 2 );

    if( i < num_items )
    {
	write( floatator_socket, &msg, sizeof(msg) );
	write( floatator_socket, &i, sizeof(int) );
    }
}

void NewCancelCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtUnmanageChild( widget.nnew );
    
    XtSetArg( args[0], XmNlistItemCount, 0 );
    XtSetArg( args[1], XmNlistItems, NULL );
    XtSetValues( widget.nnew, args, 2 );
}

void StartStopCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmToggleButtonCallbackStruct *cd = 
	(XmToggleButtonCallbackStruct *) call_data;
    int msg, success;
    
    msg = cd->set ? FDCMPStartCalc : FDCMPStopCalc;
    success = write( floatator_socket, &msg, sizeof(msg) );
}

void PaletteSelectCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    switch_off_palette();
    
    ctrl.mode = EM_SEL;
    ctrl.selected = FALSE;

    switch_on_palette();
}

void PaletteObstacleCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    switch_off_palette();
    
    ctrl.mode = EM_OBST;

    switch_on_palette();
}

void PaletteEmptyCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    switch_off_palette();
    
    ctrl.mode = EM_CLEAR;

    switch_on_palette();
}

void PaletteFluidCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    int f = *(int *)client_data;
    
    switch_off_palette();

    ctrl.mode = EM_FLUID;
    ctrl.flavor = f;

    switch_on_palette();
}

void PaletteParticleCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    int f = *(int *)client_data;
    
    switch_off_palette();

    ctrl.mode = EM_PART;
    ctrl.flavor = f;

    switch_on_palette();
}

void ViewZoomCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmScaleCallbackStruct *cd = 
	(XmScaleCallbackStruct *) call_data;
	
    view.zoom = cd->value;
    update_view();
}

void ViewColorCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmToggleButtonCallbackStruct *cd = 
	(XmToggleButtonCallbackStruct *) call_data;
    enum DisplayMode mode = (enum DisplayMode) client_data;

    if( cd->set )
    {
	view.background = mode;
	update_view();
    }
}

void ViewArrowSwitchCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmToggleButtonCallbackStruct *cd = 
	(XmToggleButtonCallbackStruct *) call_data;
	
    view.arrows = cd->set;
    update_view();
}

void ViewArrowZoomCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmScaleCallbackStruct *cd = 
	(XmScaleCallbackStruct *) call_data;
	
    view.arrow_length = cd->value;
    update_view();
}

void ViewParticlesCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmToggleButtonCallbackStruct *cd = 
	(XmToggleButtonCallbackStruct *) call_data;
	
    view.particles = cd->set;
    update_view();
}

void ViewGridCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XmToggleButtonCallbackStruct *cd = 
	(XmToggleButtonCallbackStruct *) call_data;
	
    view.grid = cd->set;
    update_view();
}

void PropertiesSetCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtUnmanageChild( widget.properties );
}

void PropertiesCancelCallback
( 
    Widget w, 
    XtPointer client_data, 
    XtPointer call_data 
)
{
    XtUnmanageChild( widget.properties );
}
