/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      creation.c                                             *
*                                                                    *
*  Contents:  Creates the windows and dialogs of the user            *
*             interface                                              *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  01.08.95 I took the code from ctrl.c             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

/********************************************************************\
*                                                                    *
*                              Includes                              *
*                                                                    *
\********************************************************************/

#include "ctrl.h"
#include "view.h"

#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MainW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ScrollBar.h>
#include <Xm/SelectioB.h>
#include <Xm/SeparatoG.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>

#include "Slider.h"

/* Pictures for the Buttons */

#include "icons/playButton.xbm"
#include "icons/pauseButton.xbm"

#include "icons/selection.xbm"
#include "icons/block.xbm"
#include "icons/particle1.xbm"
#include "icons/particle2.xbm"
#include "icons/particle3.xbm"
#include "icons/particle4.xbm"
#include "icons/particle5.xbm"
#include "icons/particle6.xbm"

/********************************************************************\
*                                                                    *
*                           Hilfsroutinen                            *
*                                                                    *
\********************************************************************/

static void CreateFileMenu()
{
    widget.Main.Menu.fileShell = XmCreatePulldownMenu
    (
        widget.Main.menu,
        WN_FILE_MENU_SHELL,
	args, 0
    );
    widget.Main.Menu.File.nnew = XmCreatePushButtonGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_NEW_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.File.nnew,
	XmNactivateCallback,
	FileNewCallback,
	NULL
    );
    widget.Main.Menu.File.sep1 = XmCreateSeparatorGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_SEP1_MENU,
	args, 0
    );
    widget.Main.Menu.File.open = XmCreatePushButtonGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_OPEN_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.File.open,
	XmNactivateCallback,
	FileOpenCallback,
	NULL
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.File.save = XmCreatePushButtonGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_SAVE_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.File.save,
	XmNactivateCallback,
	FileSaveCallback,
	NULL
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.File.saveAs = XmCreatePushButtonGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_SAVEAS_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.File.saveAs,
	XmNactivateCallback,
	FileSaveAsCallback,
	NULL
    );
    widget.Main.Menu.File.sep2 = XmCreateSeparatorGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_SEP2_MENU,
	args, 0
    );
    widget.Main.Menu.File.quit = XmCreatePushButtonGadget
    (
	widget.Main.Menu.fileShell,
        WN_FILE_QUIT_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.File.quit,
	XmNactivateCallback,
	FileQuitCallback,
	NULL
    );
    XtManageChildren( &widget.Main.Menu.File.nnew, 7 );
}

static void CreateEditMenu()
{
    widget.Main.Menu.editShell = XmCreatePulldownMenu
    (
        widget.Main.menu,
        WN_EDIT_MENU_SHELL,
	args, 0
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.Edit.cut = XmCreatePushButtonGadget
    (
	widget.Main.Menu.editShell,
        WN_EDIT_CUT_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.Edit.cut,
	XmNactivateCallback,
	EditCutCallback,
	NULL
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.Edit.copy = XmCreatePushButtonGadget
    (
	widget.Main.Menu.editShell,
        WN_EDIT_COPY_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.Edit.copy,
	XmNactivateCallback,
	EditCopyCallback,
	NULL
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.Edit.paste = XmCreatePushButtonGadget
    (
	widget.Main.Menu.editShell,
        WN_EDIT_PASTE_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.Edit.paste,
	XmNactivateCallback,
	EditPasteCallback,
	NULL
    );
    widget.Main.Menu.Edit.sep = XmCreateSeparatorGadget
    (
	widget.Main.Menu.editShell,
        WN_EDIT_SEP_MENU,
	args, 0
    );
    XtSetArg( args[0],  XmNsensitive, False );
    widget.Main.Menu.Edit.props = XmCreatePushButtonGadget
    (
	widget.Main.Menu.editShell,
        WN_EDIT_PROPS_MENU,
	args, 1
    );
    XtAddCallback
    (
        widget.Main.Menu.Edit.props,
	XmNactivateCallback,
	EditPropertiesCallback,
	NULL
    );
    XtManageChildren( &widget.Main.Menu.Edit.cut, 5 );
}

static void CreateOptionsMenu()
{
    widget.Main.Menu.optionsShell = XmCreatePulldownMenu
    (
        widget.Main.menu,
        WN_OPTIONS_MENU_SHELL,
	args, 0
    );
    widget.Main.Menu.Options.computation = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_COMPUTATION_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.computation,
	XmNactivateCallback,
	OptionsComputationCallback,
	NULL
    );
    widget.Main.Menu.Options.geometry = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_GEOMETRY_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.geometry,
	XmNactivateCallback,
	OptionsGeometryCallback,
	NULL
    );
    widget.Main.Menu.Options.time = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_TIME_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.time,
	XmNactivateCallback,
	OptionsTimeCallback,
	NULL
    );
    widget.Main.Menu.Options.pressure = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_PRESSURE_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.pressure,
	XmNactivateCallback,
	OptionsPressureCallback,
	NULL
    );
    widget.Main.Menu.Options.problem = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_PROBLEM_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.problem,
	XmNactivateCallback,
	OptionsProblemCallback,
	NULL
    );
    widget.Main.Menu.Options.boundary = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_BOUNDARY_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.boundary,
	XmNactivateCallback,
	OptionsBoundaryCallback,
	NULL
    );
    widget.Main.Menu.Options.particles = XmCreatePushButtonGadget
    (
	widget.Main.Menu.optionsShell,
        WN_OPTIONS_PARTICLES_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Options.particles,
	XmNactivateCallback,
	OptionsParticlesCallback,
	NULL
    );
    XtManageChildren( &widget.Main.Menu.Options.computation, 7 ); 
}

static void CreateHelpMenu()
{
    widget.Main.Menu.helpShell = XmCreatePulldownMenu
    (
        widget.Main.menu,
        WN_HELP_MENU_SHELL,
	args, 0
    );
    widget.Main.Menu.Help.about = XmCreatePushButtonGadget
    (
	widget.Main.Menu.helpShell,
        WN_HELP_ABOUT_MENU,
	args, 0
    );
    XtAddCallback
    (
        widget.Main.Menu.Help.about,
	XmNactivateCallback,
	HelpAboutCallback,
	NULL
    );
    XtManageChildren( &widget.Main.Menu.Help.about, 1 ); 
}

static void CreateMainMenubar() 
{
    widget.Main.menu = XmCreateMenuBar
    (
	widget.main,
        WN_MAIN_MENU,
	args, 0
    );

    CreateFileMenu();
    CreateEditMenu();
    CreateOptionsMenu();
    CreateHelpMenu();

    XtSetArg( args[0], XmNsubMenuId, widget.Main.Menu.fileShell );
    widget.Main.Menu.fileButton = XmCreateCascadeButtonGadget
    (
	widget.Main.menu,
	WN_FILE_MENU_BUTTON, 
	args, 1
    );
    XtSetArg( args[0], XmNsubMenuId, widget.Main.Menu.editShell );
    widget.Main.Menu.editButton = XmCreateCascadeButtonGadget
    (
	widget.Main.menu,
	WN_EDIT_MENU_BUTTON, 
	args, 1
    );
    XtSetArg( args[0], XmNsubMenuId, widget.Main.Menu.optionsShell );
    widget.Main.Menu.optionsButton = XmCreateCascadeButtonGadget
    (
	widget.Main.menu,
	WN_OPTIONS_MENU_BUTTON, 
	args, 1
    );
    XtSetArg( args[0], XmNsubMenuId, widget.Main.Menu.helpShell );
    widget.Main.Menu.helpButton = XmCreateCascadeButtonGadget
    (
	widget.Main.menu,
	WN_HELP_MENU_BUTTON, 
	args, 1
    );
    XtSetArg( args[0], XmNmenuHelpWidget, widget.Main.Menu.helpButton );
    XtSetValues( widget.Main.menu, args, 1 );
    
    XtManageChildren( &widget.Main.Menu.fileButton, 4 );
}

static void CreateMainView() 
{
    widget.Main.view = XmCreateDrawingArea
    (
       widget.main,
       WN_MAIN_VIEW,
       args, 0
    );
    XtAddCallback
    (
        widget.Main.view,
	XmNexposeCallback,
	UpdateViewCallback,
	NULL
    );
    XtAddCallback
    (
	widget.Main.view, 
	XmNinputCallback,
	MouseClickCallback,
	NULL
    );
}

static void CreateMainScrollbars()
{
    widget.Main.hscroll = XmCreateScrollBar
    (
       widget.main,
       WN_MAIN_HSCROLL,
       args, 0
    );
    XtAddCallback
    (
        widget.Main.hscroll,
	XmNdragCallback,
	HScrollCallback,
	NULL
    );
    XtAddCallback
    (
        widget.Main.hscroll,
	XmNvalueChangedCallback,
	HScrollCallback,
	NULL
    );
    widget.Main.vscroll = XmCreateScrollBar
    (
       widget.main,
       WN_MAIN_VSCROLL,
       args, 0
    );
    XtAddCallback
    (
        widget.Main.vscroll,
	XmNdragCallback,
	VScrollCallback,
	NULL
    );
    XtAddCallback
    (
        widget.Main.vscroll,
	XmNvalueChangedCallback,
	VScrollCallback,
	NULL
    );
}

static void CreateMainStatusline() 
{
    /* Die Initialisierung der Form-Constraint-Resourcen muss hier leider */
    /* explizit erfolgen, da der Mechanismus ueber des Resource-File      */
    /* versagt, weil die anderen Widgets bei der Erzeugung noch gar nicht */
    /* existieren.                                                        */ 

    Pixel foreground, background;
    Pixmap startButton, stopButton;

    widget.Main.statusline = XmCreateForm
    (
	widget.main,
        WN_MAIN_STATUS,
	args, 0
    );
    widget.Main.Statusline.time = XmCreateLabelGadget
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_TIME,
	args, 0
    );
    widget.Main.Statusline.state = XmCreateLabel
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_STATE,
	args, 0
    );
    widget.Main.Statusline.sep1 = XmCreateSeparatorGadget
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_SEP1,
	args, 0
    );
    widget.Main.Statusline.progres = XmCreateDrawingArea
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_PROGRES,
	args, 0
    );
    widget.Main.Statusline.sep2 = XmCreateSeparatorGadget
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_SEP2,
	args, 0
    );
    widget.Main.Statusline.residuum = XmCreateLabel
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_RESIDUUM,
	args, 0
    );
    widget.Main.Statusline.startStop = XmCreateToggleButton
    (
	widget.Main.statusline,
        WN_MAIN_STATUS_STARTSTOP,
	args, 0
    );
    
    XtSetArg( args[0], XmNforeground, &foreground );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Main.Statusline.startStop, args, 2 );
    startButton = XCreatePixmapFromBitmapData
    (
        display, DefaultRootWindow(display),
	playButton_bits, playButton_width, playButton_height,
	foreground, background,
	DefaultDepthOfScreen
        ( 
            ScreenOfDisplay
            ( 
                display, 
                DefaultScreen( display ) 
            ) 
	)
    );
    stopButton = XCreatePixmapFromBitmapData
    (
        display, DefaultRootWindow(display),
	pauseButton_bits, pauseButton_width, pauseButton_height,
	foreground, background,
	DefaultDepthOfScreen
        ( 
            ScreenOfDisplay
            ( 
                display, 
                DefaultScreen( display ) 
            ) 
	)
    );
    
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNrightAttachment, XmATTACH_NONE );
    XtSetValues( widget.Main.Statusline.time, args, 2 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNleftWidget, widget.Main.Statusline.time );
    XtSetArg( args[2], XmNrightAttachment, XmATTACH_NONE );
    XtSetValues( widget.Main.Statusline.state, args, 3 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNleftWidget, widget.Main.Statusline.state );
    XtSetArg( args[2], XmNrightAttachment, XmATTACH_NONE );
    XtSetValues( widget.Main.Statusline.sep1, args, 3 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNleftWidget, widget.Main.Statusline.sep1 );
    XtSetArg( args[2], XmNrightAttachment, XmATTACH_WIDGET );
    XtSetArg( args[3], XmNrightWidget, widget.Main.Statusline.sep2 );
    XtSetValues( widget.Main.Statusline.progres, args, 4 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_NONE );
    XtSetArg( args[1], XmNrightAttachment, XmATTACH_WIDGET );
    XtSetArg( args[2], XmNrightWidget, widget.Main.Statusline.residuum );
    XtSetValues( widget.Main.Statusline.sep2, args, 3 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_NONE );
    XtSetArg( args[1], XmNrightAttachment, XmATTACH_WIDGET );
    XtSetArg( args[2], XmNrightWidget, widget.Main.Statusline.startStop );
    XtSetValues( widget.Main.Statusline.residuum, args, 3 );
    XtSetArg( args[0], XmNleftAttachment, XmATTACH_NONE );
    XtSetArg( args[1], XmNrightAttachment, XmATTACH_FORM );
    XtSetArg( args[2], XmNlabelType, XmPIXMAP );
    XtSetArg( args[3], XmNlabelPixmap, startButton ); 
    XtSetArg( args[4], XmNselectPixmap, stopButton );
    XtSetValues( widget.Main.Statusline.startStop, args, 5 );
    
    XtManageChildren( &widget.Main.Statusline.time, 7 );
    
    XtAddCallback
    (
        widget.Main.Statusline.startStop,
	XmNvalueChangedCallback,
	StartStopCallback,
	NULL
    );
    XtAddCallback
    (
	widget.Main.Statusline.progres, 
	XmNexposeCallback, 
	ProgresCallback, 
	NULL
    );
}

static void CreateMainWindow()
{
    widget.main = XmCreateMainWindow
    (
	widget.toplevel,
        WN_MAIN,
	args, 0
    );

    CreateMainMenubar();
    CreateMainView();
    CreateMainScrollbars();
    CreateMainStatusline();

    XmMainWindowSetAreas
    (
        widget.main,
        widget.Main.menu,
	widget.Main.statusline,
	widget.Main.hscroll,
	widget.Main.vscroll,
	widget.Main.view
    );
    
    XtManageChildren( &widget.Main.menu, 5 );
}

static void CreatePaletteDialog()
{
    Pixel foreground, background;
    Pixmap pixmap;
    int i;

    widget.palette = XmCreateFormDialog
    (
        widget.toplevel,
	WN_PALETTE,
	args, 0
    );
    widget.Palette.group1 = XmCreateForm
    (
	widget.palette,
        WN_PALETTE_GROUP1,
	args, 0
    );
    widget.Palette.select = XmCreateToggleButton
    (
	widget.Palette.group1,
        WN_PALETTE_SELECT,
	args, 0
    );
    widget.Palette.obstacle = XmCreateToggleButton
    (
	widget.Palette.group1,
        WN_PALETTE_OBSTACLE,
	args, 0
    );
    widget.Palette.sep1 = XmCreateSeparatorGadget
    (
	widget.palette,
        WN_PALETTE_SEP1,
	args, 0
    );
    widget.Palette.group2 = XmCreateRowColumn
    (
	widget.palette,
        WN_PALETTE_GROUP2,
	args, 0
    );
    widget.Palette.fluids[0] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_NO_FLUID,
	args, 0
    );
    widget.Palette.fluids[1] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_FLUID_1,
	args, 0
    );
    widget.Palette.fluids[2] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_FLUID_2,
	args, 0
    );
    widget.Palette.fluids[3] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_FLUID_3,
	args, 0
    );
    widget.Palette.fluids[4] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_FLUID_4,
	args, 0
    );
    widget.Palette.fluids[5] = XmCreateToggleButton
    (
	widget.Palette.group2,
        WN_PALETTE_FLUID_5,
	args, 0
    );
    widget.Palette.sep2 = XmCreateSeparatorGadget
    (
	widget.palette,
        WN_PALETTE_SEP2,
	args, 0
    );
    widget.Palette.group3 = XmCreateRowColumn
    (
	widget.palette,
        WN_PALETTE_GROUP3,
	args, 0
    );
    widget.Palette.particles[0] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_1,
	args, 0
    );
    widget.Palette.particles[1] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_2,
	args, 0
    );
    widget.Palette.particles[2] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_3,
	args, 0
    );
    widget.Palette.particles[3] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_4,
	args, 0
    );
    widget.Palette.particles[4] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_5,
	args, 0
    );
    widget.Palette.particles[5] = XmCreateToggleButton
    (
	widget.Palette.group3,
        WN_PALETTE_PARTICLE_6,
	args, 0
    );
    
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNbottomAttachment, XmATTACH_NONE );
    XtSetValues( widget.Palette.group1, args, 2 );
    XtSetArg( args[0], XmNforeground, &foreground );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.select, args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	selection_bits, selection_width, selection_height,
	foreground, background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetArg( args[2], XmNset, True );
    XtSetArg( args[3], XmNsensitive, False );
    XtSetValues( widget.Palette.select, args, 4 );
    XtSetArg( args[0], XmNforeground, &color.obstacle );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.obstacle, args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.obstacle, background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.obstacle, args, 2 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Palette.group1 );
    XtSetArg( args[2], XmNbottomAttachment, XmATTACH_NONE );
    XtSetValues( widget.Palette.sep1, args, 3 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Palette.sep1 );
    XtSetArg( args[2], XmNbottomAttachment, XmATTACH_NONE );
    XtSetValues( widget.Palette.group2, args, 3 );
    XtSetArg( args[0], XmNforeground, &color.empty );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[0], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.empty, background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[0], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.fluid[0] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[1], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.fluid[0], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[1], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.fluid[1] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[2], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.fluid[1], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[2], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.fluid[2] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[3], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.fluid[2], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[3], args, 2   );
    XtSetArg( args[0], XmNforeground, &color.fluid[3] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[4], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.fluid[3], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[4], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.fluid[4] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.fluids[5], args, 2 );
    pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	block_bits, block_width, block_height,
	color.fluid[4], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.fluids[5], args, 2 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Palette.group2 );
    XtSetArg( args[2], XmNbottomAttachment, XmATTACH_NONE );
    XtSetValues( widget.Palette.sep2, args, 3 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Palette.sep2 );
    XtSetArg( args[2], XmNbottomAttachment, XmATTACH_NONE );
    XtSetValues( widget.Palette.group3, args, 3 );
    XtSetArg( args[0], XmNforeground, &color.particle[0] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[0], args, 2 );
    pixmaps.particle[0] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle1_bits, particle1_width, particle1_height,
	color.particle[0], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[0], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.particle[1] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[1], args, 2 );
    pixmaps.particle[1] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle2_bits, particle2_width, particle2_height,
	color.particle[1], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[1], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.particle[2] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[2], args, 2 );
    pixmaps.particle[2] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle3_bits, particle3_width, particle3_height,
	color.particle[2], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[2], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.particle[3] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[3], args, 2 );
    pixmaps.particle[3] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle4_bits, particle4_width, particle4_height,
	color.particle[3], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[3], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.particle[4] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[4], args, 2 );
    pixmaps.particle[4] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle5_bits, particle5_width, particle5_height,
	color.particle[4], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[4], args, 2 );
    XtSetArg( args[0], XmNforeground, &color.particle[5] );
    XtSetArg( args[1], XmNbackground, &background );
    XtGetValues( widget.Palette.particles[5], args, 2 );
    pixmaps.particle[5] = pixmap = XCreatePixmapFromBitmapData
    (
	display, DefaultRootWindow(display),
	particle6_bits, particle6_width, particle6_height,
	color.particle[5], background,
	DefaultDepthOfScreen
	( 
	    ScreenOfDisplay
	    ( 
		display, 
		DefaultScreen( display ) 
	    ) 
	)    
    );
    XtSetArg( args[0], XmNlabelPixmap, pixmap );
    XtSetArg( args[1], XmNselectInsensitivePixmap, pixmap );
    XtSetValues( widget.Palette.particles[5], args, 2 );
    
    XtManageChildren( &widget.Palette.select, 2 );
    XtManageChildren( widget.Palette.fluids, 6 );
    XtManageChildren( widget.Palette.particles, 6 );
    XtManageChildren( &widget.Palette.group1, 5 );
    
    XtAddCallback
    (
        widget.Palette.select,
	XmNvalueChangedCallback,
	PaletteSelectCallback,
	NULL
    );
    XtAddCallback
    (
        widget.Palette.obstacle,
	XmNvalueChangedCallback,
	PaletteObstacleCallback,
	NULL
    );
    XtAddCallback
    (
        widget.Palette.fluids[0],
	XmNvalueChangedCallback,
	PaletteEmptyCallback,
	NULL
    );
    for( i = 1; i <= FLUID_FLAVORS; i++ )
    {
	int *call_data = (int *)malloc( sizeof(int) );
	if( call_data == 0 )
	{
	    fprintf( stderr, "Out of memory.\n" );
	    exit( 1 );
	}
	*call_data = i-1;
        XtAddCallback
	(
	    widget.Palette.fluids[i],
	    XmNvalueChangedCallback,
	    PaletteFluidCallback,
	    (XtPointer) call_data
	);
    }
    for( i = 0; i < PARTICLE_FLAVORS; i++ )
    {
	int *call_data = (int *)malloc( sizeof(int) );
	if( call_data == 0 )
	{
	    fprintf( stderr, "Out of memory.\n" );
	    exit( 1 );
	}
	*call_data = i;
        XtAddCallback
	(
	    widget.Palette.particles[i],
	    XmNvalueChangedCallback,
	    PaletteParticleCallback,
	    (XtPointer) call_data
	);
    }
}

static void CreateViewDialog() 
{
    widget.view = XmCreateFormDialog
    (
        widget.toplevel,
	WN_VIEW,
	args, 0
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    widget.View.zoom_label = XmCreateLabelGadget
    (
	widget.view,
        WN_VIEW_ZOOM_LABEL,
	args, 1
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.zoom_label );
    widget.View.zoom = XmCreateScale
    (
	widget.view,
        WN_VIEW_ZOOM,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.zoom );
    widget.View.sep1 = XmCreateSeparatorGadget
    (
	widget.view,
        WN_VIEW_ZOOM,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.zoom );
    widget.View.color_label = XmCreateLabelGadget
    (
	widget.view,
        WN_VIEW_COLOR_LABEL,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.color_label );
    widget.View.color = XmCreateRowColumn
    (
	widget.view,
        WN_VIEW_COLOR,
	args, 2
    );
    XtSetArg( args[0], XmNset, True );
    widget.View.Color.u = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_U,
	args, 1
    );
    widget.View.Color.v = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_V,
	args, 0
    );
    widget.View.Color.p = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_P,
	args, 0
    );
    widget.View.Color.psi = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_PSI,
	args, 0
    );
    widget.View.Color.zeta = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_ZETA,
	args, 0
    );
    widget.View.Color.fluids = XmCreateToggleButtonGadget
    (
	widget.View.color,
        WN_VIEW_COLOR_FLUIDS,
	args, 0
    );
    XtManageChildren( &widget.View.Color.u, 6 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.color );
    widget.View.sep2 = XmCreateSeparatorGadget
    (
	widget.view,
        WN_VIEW_ZOOM,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.sep2 );
    widget.View.arrow_switch = XmCreateToggleButtonGadget
    (
	widget.view,
        WN_VIEW_ARROW_SWITCH,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.arrow_switch );
    widget.View.arrow_zoom = XmCreateScale
    (
	widget.view,
        WN_VIEW_ARROW_ZOOM,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.arrow_zoom );
    widget.View.sep3 = XmCreateSeparatorGadget
    (
	widget.view,
        WN_VIEW_SEP3,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.sep3 );
    XtSetArg( args[2], XmNset, True );
    widget.View.particles = XmCreateToggleButtonGadget
    (
	widget.view,
        WN_VIEW_PARTICLES,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.particles );
    widget.View.sep4 = XmCreateSeparatorGadget
    (
	widget.view,
        WN_VIEW_SEP4,
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.View.sep4 );
    widget.View.grid = XmCreateToggleButtonGadget
    (
	widget.view,
        WN_VIEW_GRID,
	args, 2
    );

    XtManageChildren( &widget.View.zoom, 12 );
    
    XtAddCallback
    (
	widget.View.zoom, 
	XmNdragCallback, 
	ViewZoomCallback, 
	NULL
    );
    XtAddCallback
    (
	widget.View.Color.u, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_U
    );
    XtAddCallback
    (
	widget.View.Color.v, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_V
    );
    XtAddCallback
    (
	widget.View.Color.p, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_P
    );
    XtAddCallback
    (
	widget.View.Color.fluids, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_FLUID
    );
    XtAddCallback
    (
	widget.View.Color.psi, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_PSI
    );
    XtAddCallback
    (
	widget.View.Color.zeta, 
	XmNvalueChangedCallback, 
	ViewColorCallback, 
	(XtPointer) DM_ZETA
    );
    XtAddCallback
    (
	widget.View.arrow_switch, 
	XmNvalueChangedCallback, 
	ViewArrowSwitchCallback, 
	NULL
    );
    XtAddCallback
    (
	widget.View.arrow_zoom, 
	XmNdragCallback, 
	ViewArrowZoomCallback, 
	NULL
    );
    XtAddCallback
    (
	widget.View.particles, 
	XmNvalueChangedCallback, 
	ViewParticlesCallback, 
	NULL
    );
    XtAddCallback
    (
	widget.View.grid, 
	XmNvalueChangedCallback, 
	ViewGridCallback, 
	NULL
    );
}

static void CreateFileNewDialog() 
{
    Widget w;
    widget.nnew = XmCreateSelectionDialog
    (
        widget.toplevel,
	WN_NEW,
	args, 0
    );
    w = XmSelectionBoxGetChild
    (
	widget.nnew, 
	XmDIALOG_APPLY_BUTTON
    );
    XtUnmanageChild( w );
    w = XmSelectionBoxGetChild
    (
	widget.nnew, 
	XmDIALOG_HELP_BUTTON
    );
    XtUnmanageChild( w );
    XtAddCallback
    (
	widget.nnew, 
	XmNokCallback, 
	NewOKCallback, 
	NULL
    );
    XtAddCallback
    (
	widget.nnew, 
	XmNcancelCallback, 
	NewCancelCallback, 
	NULL
    );
}

static void CreateFileOpenDialog()
{
    widget.open = XmCreateFileSelectionDialog
    (
        widget.toplevel,
	WN_OPEN,
	args, 0
    );
}

static void CreateFileSaveAsDialog()
{
    widget.saveAs = XmCreateFileSelectionDialog
    (
        widget.toplevel,
	WN_SAVEAS,
	args, 0
    );
}

static void CreateEditPropertiesDialog()
{
    widget.properties = XmCreateFormDialog
    (
        widget.toplevel,
	WN_PROPERTIES,
	args, 0
    );
    widget.Properties.group = XmCreateRowColumn
    (
	widget.properties,
        WN_PROPERTIES_GROUP,
	args, 0
    );
    widget.Properties.u_label = XmCreateLabelGadget
    (
        widget.Properties.group,
	WN_PROPERTIES_U_LABEL,
	args, 0
    );
    widget.Properties.v_label = XmCreateLabelGadget
    (
        widget.Properties.group,
	WN_PROPERTIES_V_LABEL,
	args, 0
    );
    widget.Properties.p_label = XmCreateLabelGadget
    (
        widget.Properties.group,
	WN_PROPERTIES_P_LABEL,
	args, 0
    );
    widget.Properties.u = XmCreateTextField
    (
        widget.Properties.group,
	WN_PROPERTIES_U,
	args, 0  
    );
    widget.Properties.v = XmCreateTextField
    (
        widget.Properties.group,
	WN_PROPERTIES_V,
	args, 0  
    );
    widget.Properties.p = XmCreateTextField
    (
        widget.Properties.group,
	WN_PROPERTIES_P,
	args, 0  
    );
    widget.Properties.sep1 = XmCreateSeparatorGadget
    (
	widget.properties,
        WN_PROPERTIES_SEP1,
	args, 0
    );
    widget.Properties.set = XmCreatePushButtonGadget
    (
	widget.properties,
        WN_PROPERTIES_SET,
	args, 0
    );
    widget.Properties.cancel = XmCreatePushButtonGadget
    (
	widget.properties,
        WN_PROPERTIES_CANCEL,
	args, 0
    );

    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNleftAttachment, XmATTACH_FORM );
    XtSetValues( widget.Properties.group, args, 2 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Properties.group );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    XtSetValues( widget.Properties.sep1, args, 4 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Properties.sep1 );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetValues( widget.Properties.set, args, 3 );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Properties.sep1 );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_NONE );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    XtSetValues( widget.Properties.cancel, args, 4 );

    XtManageChildren( &widget.Properties.u, 6 );
    XtManageChildren( &widget.Properties.set, 4 );

    XtAddCallback
    (
        widget.Properties.set,
	XmNactivateCallback,
	PropertiesSetCallback,
	NULL
    );
    XtAddCallback
    (
        widget.Properties.cancel,
	XmNactivateCallback,
	PropertiesCancelCallback,
	NULL
    );
}

static void CreateOptionsComputationDialog()
{
    widget.computation = XmCreateFormDialog
    (
        widget.toplevel,
	WN_COMPUTATION,
	args, 0
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNleftAttachment, XmATTACH_FORM );
    widget.Computation.server_label = XmCreateLabelGadget
    (
	widget.computation, 
	WN_COMPUTATION_SERVER_LABEL, 
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.connect = XmCreatePushButtonGadget
    (
	widget.computation, 
	WN_COMPUTATION_CONNECT_BUTTON, 
	args, 2
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.server_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    widget.Computation.interface_label = XmCreateLabelGadget
    (
	widget.computation, 
	WN_COMPUTATION_INTERFACE_LABEL, 
	args, 3
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.interface_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[3], XmNleftWidget, widget.Computation.interface_label );
    XtSetArg( args[4], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.interface = XmCreateTextField
    (
	widget.computation, 
	WN_COMPUTATION_INTERFACE, 
	args, 5
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.interface_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    widget.Computation.socket_label = XmCreateLabelGadget
    (
	widget.computation, 
	WN_COMPUTATION_SOCKET_LABEL, 
	args, 3
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.socket_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_WIDGET );
    XtSetArg( args[3], XmNleftWidget, widget.Computation.socket_label );
    XtSetArg( args[4], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.socket = XmCreateTextField
    (
	widget.computation, 
	WN_COMPUTATION_SOCKET, 
	args, 5
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.socket_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.sep1 = XmCreateSeparatorGadget
    (
	widget.computation, 
	WN_COMPUTATION_SEP1, 
	args, 4
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.sep1 );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.iproc_label = XmCreateLabelGadget
    (
	widget.computation, 
	WN_COMPUTATION_IPROC_LABEL, 
	args, 4
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.iproc_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.iproc = CreateSlider
    (
	widget.computation, 
	WN_COMPUTATION_IPROC, 
	args, 4, 
	NULL, NULL
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.iproc );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.jproc_label = XmCreateLabelGadget
    (
	widget.computation, 
	WN_COMPUTATION_JPROC_LABEL, 
	args, 4
    );
    XtSetArg( args[0], XmNtopAttachment, XmATTACH_WIDGET );
    XtSetArg( args[1], XmNtopWidget, widget.Computation.jproc_label );
    XtSetArg( args[2], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNrightAttachment, XmATTACH_FORM );
    widget.Computation.jproc = CreateSlider
    (
	widget.computation, 
	WN_COMPUTATION_JPROC, 
	args, 4, 
	NULL, NULL
    );
    
    XtManageChildren( &widget.Computation.server_label, 11 );
}

static void CreateOptionsGeometryDialog() {}
static void CreateOptionsTimeDialog() {}
static void CreateOptionsPressureDialog() {}
static void CreateOptionsProblemDialog() {}
static void CreateOptionsBoundaryDialog() {}
static void CreateOptionsParticlesDialog() {}

static char important_message [] = 
"Floatator - interactive fluid dynamics\n\
For license and other information, please visit the homepage: \n\
http://www.cichon.com/floatator/\n\
\n\
Dear unnamed person, \nIt was so great being with you!\nGordon.\n";

static void CreateAboutDialog()
{
    widget.about = XmCreateBulletinBoardDialog
    (
        widget.toplevel,
	WN_ABOUT,
	args, 0	
    );

    XtSetArg( args[0], XmNtopAttachment, XmATTACH_FORM );
    XtSetArg( args[1], XmNleftAttachment, XmATTACH_FORM );
    XtSetArg( args[2], XmNrightAttachment, XmATTACH_FORM );
    XtSetArg( args[3], XmNbottomAttachment, XmATTACH_FORM );
    XtSetArg( args[4], XmNeditable, False );
    XtSetArg( args[5], XmNeditMode, XmMULTI_LINE_EDIT );
    XtSetArg( args[6], XmNvalue, important_message );
    XtSetArg( args[7], XmNwordWrap, True);
    XtSetArg( args[8], XmNrows, 10);
    XtSetArg( args[9], XmNcolumns, 40);
    widget.About.text = XmCreateText
    (
	widget.about, 
	WN_ABOUT_TEXT, 
	args, 10
    );
    
    XtManageChildren( &widget.About.text, 1 );
}



/* Zeigt gebrauchsinformationen zum Programm an */
static void usage( FILE *f )
{
    fprintf( f, "\n" );
    fprintf( f, "  The Floatator\n" );
    fprintf( f, "                 by Gordon Cichon 1995\n" );
    fprintf( f, "\n" );
    fprintf( f, "  Usage: floatator [filename]\n" ); 
    fprintf( f, "\n" );
}

/********************************************************************\
*                                                                    *
*                             Funktionen                             *
*                                                                    *
\********************************************************************/

void ctrl_init( int *argc, char **argv )
{
    open_X_connection( argc, argv );
    parse_commandline( argc, argv );
    create_windows();
    init_cursor_pixmaps();
}

void open_X_connection( int *argc, char **argv )
{
    XtToolkitInitialize();
    context = XtCreateApplicationContext();
    display = XtOpenDisplay
    (
        context,           /* application context */
	NULL,              /* use DISPLAY from environment */
	NULL,              /* use last of argv[0] as name */
	PROGRAM_CLASS,
	NULL, 0,           /* additional command line options */
	argc, argv         /* use and delete standard options */
    );
    if( display == NULL )
        XtErrorMsg
	(
            NErrCannotOpen,
	    TErrMain,
	    ERROR_CLASS,
	    "cannot open display",
	    NULL, NULL
        );
    XtSetArg( args[0], XmNtitle, MAIN_TITLE );
    widget.toplevel = XtAppCreateShell
    (
        NULL,             /* Use the same name */
	PROGRAM_CLASS,
	applicationShellWidgetClass,
	display,
	args, 1
    );
}

/* Analysiert die Kommandozeile */
void parse_commandline( int *argc, char **argv )
{
    int c = *argc;
    switch( c )
    {
      case 1:
	strcpy( filename, NO_FILE_LOADED );
	break;
      case 2:
	if( strcasecmp( "-h", argv[1] ) == 0 
	|| strcasecmp( "-?", argv[1] ) == 0
	|| strcasecmp( "-help", argv[1] ) == 0 )
	{
	    usage( stdout );
	    exit(0);
	}
	strcpy( filename, argv[1] );
	break;
      default:
	usage( stderr ); 
	exit(1);
    }
    socket_name = strdup( SOCKET_NAME );
}

void create_windows()
{
    CreateMainWindow();

    CreatePaletteDialog();
    CreateViewDialog();

    CreateFileNewDialog();
    CreateFileOpenDialog();
    CreateFileSaveAsDialog();

    CreateEditPropertiesDialog();

    CreateOptionsComputationDialog();
    CreateOptionsGeometryDialog();
    CreateOptionsTimeDialog();
    CreateOptionsPressureDialog();
    CreateOptionsProblemDialog();
    CreateOptionsBoundaryDialog();
    CreateOptionsParticlesDialog();

    CreateAboutDialog();

    XtManageChild( widget.main );
    XtManageChild( widget.palette );
    XtManageChild( widget.view );
    XtManageChild( widget.about );
}

void init_cursor_pixmaps()
{
    cursor_pixmap.select = None;
    cursor_pixmap.obstacle = 
    cursor_pixmap.clear = 
    cursor_pixmap.fluid[0] =
    cursor_pixmap.fluid[1] =
    cursor_pixmap.fluid[2] =
    cursor_pixmap.fluid[3] =
    cursor_pixmap.fluid[4] =
    XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	block_bits, 
	block_width, block_height 
    );
    cursor_pixmap.particle[0] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle1_bits, 
	particle1_width, particle1_height 
    );
    cursor_pixmap.particle[1] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle2_bits, 
	particle2_width, particle2_height 
    );
    cursor_pixmap.particle[2] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle3_bits, 
	particle3_width, particle3_height 
    );
    cursor_pixmap.particle[3] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle4_bits, 
	particle4_width, particle4_height 
    );
    cursor_pixmap.particle[4] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle5_bits, 
	particle5_width, particle5_height 
    );
    cursor_pixmap.particle[5] = XCreateBitmapFromData
    (
	display,
	DefaultRootWindow(display), 
	particle6_bits, 
	particle6_width, particle6_height 
    );

}
