/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      ctrl.c                                                 *
*                                                                    *
*  Contents:  The implementation of the manipulation specific part   *
*             of the floatator program.                              *
*             This part of the program is concerned with             *
*             maniulating the scene by use of Xmoti.                 *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  01.08.95 First steps                             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

/********************************************************************\
*                                                                    *
*                              Includes                              *
*                                                                    *
\********************************************************************/

#include "../glib.h"

#include "ctrl.h"

#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/MainW.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>

/********************************************************************\
*                                                                    *
*                             Funktionen                             *
*                                                                    *
\********************************************************************/

void main_loop()
{
    XtAppAddInput
    ( 
        context, 
	floatator_socket,
	(XtPointer) XtInputReadMask, 
	FloatatorCallback, 
	NULL
    );
     
    XtRealizeWidget( widget.toplevel );
    
    XtAppMainLoop( context );     /* Hier versickert der Kontrollfluss */
}

void ctrl_cleanup()
{
}

int everything_saved()
{
    return 1;
}

void switch_off_palette()
{
    Widget button;
    
    switch( ctrl.mode )
    {
	case EM_SEL:
	    button = widget.Palette.select;
	    break;
	case EM_FLUID:
	    button = widget.Palette.fluids[ctrl.flavor+1]; 
	    break;
	case EM_PART:
	    button = widget.Palette.particles[ctrl.flavor];
	    break;
	case EM_OBST:
	    button = widget.Palette.obstacle;
	    break;
	case EM_CLEAR:
	    button = widget.Palette.fluids[0]; 
	    break;
    }
    
    XmToggleButtonSetState( button, False, False );
    XtSetSensitive( button, True );

    if( allocated_cursor != None )
    {
	XUndefineCursor( display, XtWindow( widget.Main.view ) );
	XFreeCursor( display, allocated_cursor );
	allocated_cursor = None;
    }  
}   

void switch_on_palette()
{
    Widget button;
    Pixmap pixmap;
    XColor foreground, background;
    Colormap colormap;
    
    switch( ctrl.mode )
    {
	case EM_SEL:
	    button = widget.Palette.select;
	    pixmap = cursor_pixmap.select;
	    break;
	case EM_FLUID:
	    button = widget.Palette.fluids[ctrl.flavor+1]; 
	    pixmap = cursor_pixmap.fluid[ctrl.flavor];
	    break;
	case EM_PART:
	    button = widget.Palette.particles[ctrl.flavor];
	    pixmap = cursor_pixmap.particle[ctrl.flavor];
	    break;
	case EM_OBST:
	    button = widget.Palette.obstacle;
	    pixmap = cursor_pixmap.obstacle;
	    break;
	case EM_CLEAR:
	    button = widget.Palette.fluids[0]; 
	    pixmap = cursor_pixmap.clear;
	    break;
    }
    
    XmToggleButtonSetState( button, True, False );
    XtSetSensitive( button, False );
    
    if( pixmap != None )
    {
	XtSetArg( args[0], XmNbackground, &background.pixel );
	XtSetArg( args[1], XmNforeground, &foreground.pixel );
	XtSetArg( args[2], XmNcolormap, &colormap );
	XtGetValues( button, args, 3 );
	XQueryColor( display, colormap, &background );
	XQueryColor( display, colormap, &foreground );
	allocated_cursor = XCreatePixmapCursor
	(
	    display, 
	    pixmap, 
	    pixmap, 
	    &foreground, 
	    &background, 
	    16, 16
	);
	XDefineCursor
	( 
	    display, 
	    XtWindow( widget.Main.view ), 
	    allocated_cursor
	);
    }
}  

XmString fetch_state_string()
{
    static int initialized = FALSE;
    static XmString strings[CS_NUMSTATES];
    
    if( !initialized )
    {
	int i;
	for( i = 0; i < CS_NUMSTATES; i++ )
	{
	    strings[i] = XmStringCreateLtoR
	    (
		calculation_state_strings[i], 
		XmFONTLIST_DEFAULT_TAG
	    );
	    if( strings[i] == NULL )
	    {
		fprintf
		( 
		    stderr, 
		    "String \"%s\" could not be allocated.\n", 
		    calculation_state_strings[i]
		);
		exit( 1 );
	    }
	}
	initialized = TRUE;
    }
    return strings[model.state];
}

Pixel fetch_state_color()
{
    static int initialized = FALSE;
    static Pixel colors[CS_NUMSTATES];
    
    if( !initialized )
    {
	int i, success;
	XrmValue from, to;
	for( i = 0; i < CS_NUMSTATES; i++ )
	{
	    from.addr = calculation_state_colors[i];
	    from.size = strlen( from.addr );
	    to.size = sizeof(Pixel);
	    success = XtConvertAndStore
	    (
	        widget.Main.Statusline.state, 
		XmRString, 
		&from, 
		XmRPixel, 
		&to
	    );
	    if( ! success )
	    {
		fprintf
		( 
		    stderr, 
		    "Color \"%s\" could not be allocated.\n", 
		    calculation_state_colors[i]
		);
		exit( 1 );
	    }
	    colors[i] = *(Pixel *)to.addr;
	}
	initialized = TRUE;
    }
    return colors[model.state];
}

Pixel fetch_residuum_color()
{
    static int initialized = FALSE;
    static Pixel good, bad;
    
    if( !initialized )
    {
	int success;
	XrmValue from, to;
	
	from.addr = residuum_colors[0];
	from.size = strlen( from.addr );
	to.size = sizeof(Pixel);
	success = XtConvertAndStore
	(
	    widget.Main.Statusline.residuum, 
	    XmRString, 
	    &from, 
	    XmRPixel, 
	    &to
	);
	if( ! success )
	{
	    fprintf
	    ( 
		stderr, 
		"Color \"%s\" could not be allocated.\n", 
		residuum_colors[0]
	    );
	    exit( 1 );
	}
	good = *(Pixel *)to.addr;

	from.addr = residuum_colors[1];
	from.size = strlen( from.addr );
	to.size = sizeof(Pixel);
	success = XtConvertAndStore
	(
	    widget.Main.Statusline.residuum, 
	    XmRString, 
	    &from, 
	    XmRPixel, 
	    &to
	);
	if( ! success )
	{
	    fprintf
	    ( 
		stderr, 
		"Color \"%s\" could not be allocated.\n", 
		residuum_colors[1]
	    );
	    exit( 1 );
	}
	bad = *(Pixel *)to.addr;

	initialized = TRUE;
    }

    return model.res <= model.parms.eps ? good : bad;
}
