/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      ctrl.h                                                 *
*                                                                    *
*  Contents:  The definition of the manipulation specific part       *
*             of the floatator program.                              *
*             This part of the program is concerned with             *
*             maniulating the scene by use of Xmoti.                 *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  01.08.95 First steps                             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

#ifndef CONTROLLER_H
#define CONTROLLER_H

#include "../glib.h"

#include "../problem/model.h"

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>

/********************************************************************\
*                                                                    *
*                         Enum-Konstanten                            *
*                         ===============                            *
*                                                                    *
*  Comment: In alphabetischer Reihenfolge                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von EditMode                         *
*                                                                    *
\********************************************************************/

enum EditMode
{
    EM_SEL,	/* Selection-Mode */
    EM_FLUID,	/* Einbringung von Fluid */
    EM_PART,	/* Einbringung von Partikeln */
    EM_OBST,	/* Einbringung von Hindernisssen */
    EM_CLEAR    /* Leeren des Gebietes */
};

#define NO_FILE_LOADED "[unbenannt]"

/********************************************************************\
*                                                                    *
*                         Datenstrukturen                            *
*                         ===============                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von CtrlParams                       *
*                                                                    *
\********************************************************************/

struct CtrlParms
{
    int selected;	    /* Ist ein Gebiet markiert                       */
    REAL xmin, xmax, ymin, ymax;
                            /* Ausmasse des markierten Gebiets               */
    REAL fluid_density;	    /* Partikeldichte des Fluids                     */
    REAL particle_density;  /* Partikeldichte fuer Streich- oder Bahnlinien  */
    REAL u, v, p;	    /* Parameter zum setzen von Gebietsanfangswerten */
    enum EditMode mode;	    /* Ausgesuchter Modus auf der Palette            */
    int  flavor;	    /* Spielart des Partikels oder Fluids            */
    enum ParticleMode pm;   /* Art der zu erzeugenden Partikel               */
};

struct CursorPixmaps
{
    Pixmap select;
    Pixmap obstacle;
    Pixmap clear;
    Pixmap fluid[FLUID_FLAVORS];
    Pixmap particle[PARTICLE_FLAVORS];
};

struct WidgetIDs
{
    Widget toplevel;

    Widget main;
    struct {
        Widget menu;
        Widget view;
        Widget hscroll, vscroll;
        Widget statusline;
        struct {
	    Widget fileButton;
	    Widget editButton;
	    Widget optionsButton;
	    Widget helpButton;
	    Widget fileShell;
	    struct {
	        Widget nnew;
	        Widget sep1;
	        Widget open;
	        Widget save;
	        Widget saveAs;
	        Widget sep2;
	        Widget quit;
	    } File;
	    Widget editShell;
	    struct {
	        Widget cut;
	        Widget copy;
	        Widget paste;
	        Widget sep;
	        Widget props;
	    } Edit;
	    Widget optionsShell;
	    struct {
	        Widget computation;
	        Widget geometry;
	        Widget time;
	        Widget pressure;
	        Widget problem;
	        Widget boundary;
	        Widget particles;
	    } Options;
	    Widget helpShell;
	    struct {
	        Widget about;
	    } Help;
	} Menu;
        struct {
	    Widget time;
	    Widget state;
	    Widget sep1;
	    Widget progres;
	    Widget sep2;
	    Widget residuum;
	    Widget startStop;
	} Statusline;
    } Main;

    Widget palette;
    struct {
        Widget select;
        Widget obstacle;
        Widget fluids[FLUID_FLAVORS+1];
        Widget particles[PARTICLE_FLAVORS];
        Widget group1, group2, group3;
        Widget sep1, sep2;
    } Palette;
    Widget view;
    struct {
        Widget zoom, zoom_label;
        Widget color, color_label;
        Widget arrow_switch, arrow_zoom;
        Widget particles, grid;
        Widget sep1, sep2, sep3, sep4;
        struct {
	    Widget u, v;
	    Widget p, fluids;
	    Widget psi, zeta;
	} Color;
    } View;

    Widget nnew;
    Widget open;
    Widget saveAs;

    Widget properties;
    struct {
        Widget u, u_label;
        Widget v, v_label;
        Widget p, p_label;
        Widget set, cancel;
        Widget sep1, group;
    } Properties;

    Widget computation;
    struct {
	Widget server_label, connect;
	Widget interface_label, interface;
	Widget socket_label, socket;
	Widget sep1;
	Widget iproc, iproc_label;
	Widget jproc, jproc_label;
    } Computation;
    
    Widget geometry;
    Widget time;
    Widget pressure;
    Widget problem;
    Widget boundary;
    Widget particles;

    Widget about;
    struct {
	Widget text;
    } About;
};

/********************************************************************\
*                                                                    *
*                        Namen der Widgets                           *
*                        =================                           *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

#define WN_TOPLEVEL                   "toplevel"
#define WN_MAIN                       "main"
#define WN_MAIN_MENU                  "menu"
#define WN_MAIN_VIEW                  "view"
#define WN_FILE_MENU_BUTTON           "fileButton"
#define WN_FILE_MENU_SHELL            "file"
#define WN_FILE_NEW_MENU              "new"
#define WN_FILE_SEP1_MENU             "sep1"
#define WN_FILE_OPEN_MENU             "open"
#define WN_FILE_SAVE_MENU             "save"
#define WN_FILE_SAVEAS_MENU           "saveAs"
#define WN_FILE_SEP2_MENU             "sep2"
#define WN_FILE_QUIT_MENU             "quit"
#define WN_EDIT_MENU_BUTTON           "editButton"
#define WN_EDIT_MENU_SHELL            "edit"
#define WN_EDIT_CUT_MENU              "cut"
#define WN_EDIT_COPY_MENU             "copy"
#define WN_EDIT_PASTE_MENU            "paste"
#define WN_EDIT_SEP_MENU              "sep"
#define WN_EDIT_PROPS_MENU            "props"
#define WN_OPTIONS_MENU_BUTTON        "optionsButton"
#define WN_OPTIONS_MENU_SHELL         "options"
#define WN_OPTIONS_COMPUTATION_MENU   "computation"
#define WN_OPTIONS_GEOMETRY_MENU      "geometry"
#define WN_OPTIONS_TIME_MENU          "time"
#define WN_OPTIONS_PRESSURE_MENU      "pressure"
#define WN_OPTIONS_PROBLEM_MENU       "problem"
#define WN_OPTIONS_BOUNDARY_MENU      "boundary"
#define WN_OPTIONS_PARTICLES_MENU     "particles"
#define WN_HELP_MENU_BUTTON           "helpButton"
#define WN_HELP_MENU_SHELL            "help"
#define WN_HELP_ABOUT_MENU            "about"
#define WN_MAIN_HSCROLL               "hscroll"
#define WN_MAIN_VSCROLL               "vscroll"
#define WN_MAIN_STATUS                "status"
#define WN_MAIN_STATUS_TIME           "time"
#define WN_MAIN_STATUS_STATE          "state"
#define WN_MAIN_STATUS_SEP1           "sep1"
#define WN_MAIN_STATUS_PROGRES        "progres"
#define WN_MAIN_STATUS_SEP2           "sep2"
#define WN_MAIN_STATUS_RESIDUUM       "residuum"
#define WN_MAIN_STATUS_STARTSTOP      "startStop"
#define WN_PALETTE                    "palette"
#define WN_PALETTE_SELECT             "select"
#define WN_PALETTE_OBSTACLE           "obstacle"
#define WN_PALETTE_NO_FLUID           "no_fluid"
#define WN_PALETTE_FLUID_1            "fluid1"
#define WN_PALETTE_FLUID_2            "fluid2"
#define WN_PALETTE_FLUID_3            "fluid3"
#define WN_PALETTE_FLUID_4            "fluid4"
#define WN_PALETTE_FLUID_5            "fluid5"
#define WN_PALETTE_PARTICLE_1         "particle1"
#define WN_PALETTE_PARTICLE_2         "particle2"
#define WN_PALETTE_PARTICLE_3         "particle3"
#define WN_PALETTE_PARTICLE_4         "particle4"
#define WN_PALETTE_PARTICLE_5         "particle5"
#define WN_PALETTE_PARTICLE_6         "particle6"
#define WN_PALETTE_GROUP1             "group1"
#define WN_PALETTE_GROUP2             "group2"
#define WN_PALETTE_GROUP3             "group3"
#define WN_PALETTE_SEP1               "sep1"
#define WN_PALETTE_SEP2               "sep2"
#define WN_VIEW                       "view"
#define WN_VIEW_ZOOM                  "zoom"
#define WN_VIEW_ZOOM_LABEL            "zoomLabel"
#define WN_VIEW_SEP1                  "sep1"
#define WN_VIEW_COLOR_LABEL           "colorLabel"
#define WN_VIEW_COLOR                 "color"
#define WN_VIEW_COLOR_U               "u"
#define WN_VIEW_COLOR_V               "v"
#define WN_VIEW_COLOR_P               "p"
#define WN_VIEW_COLOR_PSI             "psi"
#define WN_VIEW_COLOR_ZETA            "zeta"
#define WN_VIEW_COLOR_FLUIDS          "fluids"
#define WN_VIEW_SEP2                  "sep2"
#define WN_VIEW_ARROW_SWITCH          "arrowSwitch"
#define WN_VIEW_ARROW_ZOOM            "arrowZoom"
#define WN_VIEW_SEP3                  "sep3"
#define WN_VIEW_GRID                  "grid"
#define WN_VIEW_SEP4                  "sep4"
#define WN_VIEW_PARTICLES             "particles"
#define WN_NEW                        "new"
#define WN_OPEN                       "open"
#define WN_SAVEAS                     "saveAs"
#define WN_PROPERTIES                 "editProperties"
#define WN_PROPERTIES_U               "u"
#define WN_PROPERTIES_V               "v"
#define WN_PROPERTIES_P               "p"
#define WN_PROPERTIES_U_LABEL         "u_label"
#define WN_PROPERTIES_V_LABEL         "v_label"
#define WN_PROPERTIES_P_LABEL         "p_label"
#define WN_PROPERTIES_SET             "set"
#define WN_PROPERTIES_CANCEL          "cancel"
#define WN_PROPERTIES_SEP1            "sep1"
#define WN_PROPERTIES_GROUP           "group"
#define WN_COMPUTATION                "computation"
#define WN_COMPUTATION_SERVER_LABEL   "serverLabel"
#define WN_COMPUTATION_CONNECT_BUTTON "connect"
#define WN_COMPUTATION_INTERFACE_LABEL "interfaceLabel"
#define WN_COMPUTATION_INTERFACE      "interface"
#define WN_COMPUTATION_SOCKET_LABEL   "socketLabel"
#define WN_COMPUTATION_SOCKET         "socket"
#define WN_COMPUTATION_SEP1           "sep1"
#define WN_COMPUTATION_IPROC_LABEL    "iproc_label"
#define WN_COMPUTATION_IPROC          "iproc"
#define WN_COMPUTATION_JPROC_LABEL    "jproc_label"
#define WN_COMPUTATION_JPROC          "jproc"
#define WN_ABOUT                      "about"
#define WN_ABOUT_TEXT                 "text"


/********************************************************************\
*                                                                    *
*                         Motif sonstiges                            *
*                         ===============                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

#define PROGRAM_CLASS   "Floatator"
#define ERROR_CLASS     "FloatatorError"

#define NErrCannotOpen  "cannotOpen"
#define TErrMain        "Main"

/********************************************************************\
*                                                                    *
*                        globale Variablen                           *
*                        =================                           *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

extern char filename[FILENAME_MAX];  /* Dateiname der geladenen Datei */

#define NUM_ARGS 30

extern Arg          args[NUM_ARGS];
extern XmString     xms;
extern char         buffer[256];

extern Display      *display;
extern XtAppContext context;

extern struct WidgetIDs widget;

extern int floatator_socket;    /* Verbindungssocket zum Floatator */
extern char *socket_name;

extern struct CtrlParms ctrl;
extern struct ViewParms view;
extern struct DataPackage model;

extern Cursor allocated_cursor;
extern struct CursorPixmaps cursor_pixmap;

/********************************************************************\
*                                                                    *
*                          Beschriftungen                            *
*                          ==============                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  02.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

#define MAIN_TITLE      "Floatator"
extern char *calculation_state_strings[CS_NUMSTATES]; 
extern char *calculation_state_colors[CS_NUMSTATES];
extern char *residuum_colors[2];

/********************************************************************\
*                                                                    *
*                            Callbacks                               *
*                            =========                               *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  27.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

void FloatatorCallback( XtPointer, int*, XtInputId* );
void FileNewCallback( Widget, XtPointer, XtPointer );
void FileOpenCallback( Widget, XtPointer, XtPointer );
void FileSaveCallback( Widget, XtPointer, XtPointer );
void FileSaveAsCallback( Widget, XtPointer, XtPointer );
void FileQuitCallback( Widget, XtPointer, XtPointer );
void EditCutCallback( Widget, XtPointer, XtPointer );
void EditCopyCallback( Widget, XtPointer, XtPointer );
void EditPasteCallback( Widget, XtPointer, XtPointer );
void EditPropertiesCallback( Widget, XtPointer, XtPointer );
void OptionsComputationCallback( Widget, XtPointer, XtPointer );
void OptionsGeometryCallback( Widget, XtPointer, XtPointer );
void OptionsTimeCallback( Widget, XtPointer, XtPointer );
void OptionsPressureCallback( Widget, XtPointer, XtPointer );
void OptionsProblemCallback( Widget, XtPointer, XtPointer );
void OptionsBoundaryCallback( Widget, XtPointer, XtPointer );
void OptionsParticlesCallback( Widget, XtPointer, XtPointer );
void HelpAboutCallback( Widget, XtPointer, XtPointer );
void UpdateViewCallback( Widget, XtPointer, XtPointer );
void MouseClickCallback( Widget, XtPointer, XtPointer );
void HScrollCallback( Widget, XtPointer, XtPointer );
void VScrollCallback( Widget, XtPointer, XtPointer );
void ProgresCallback( Widget, XtPointer, XtPointer );
void StartStopCallback( Widget, XtPointer, XtPointer );
void PaletteSelectCallback( Widget, XtPointer, XtPointer );
void PaletteObstacleCallback( Widget, XtPointer, XtPointer );
void PaletteEmptyCallback( Widget, XtPointer, XtPointer );
void PaletteFluidCallback( Widget, XtPointer, XtPointer );
void PaletteParticleCallback( Widget, XtPointer, XtPointer );
void ViewZoomCallback( Widget, XtPointer, XtPointer );
void ViewColorCallback( Widget, XtPointer, XtPointer );
void ViewArrowSwitchCallback( Widget, XtPointer, XtPointer );
void ViewArrowZoomCallback( Widget, XtPointer, XtPointer );
void ViewParticlesCallback( Widget, XtPointer, XtPointer );
void ViewGridCallback( Widget, XtPointer, XtPointer );
void NewOKCallback( Widget, XtPointer, XtPointer );
void NewCancelCallback( Widget, XtPointer, XtPointer );
void PropertiesSetCallback( Widget, XtPointer, XtPointer );
void PropertiesCancelCallback( Widget, XtPointer, XtPointer );
void ComputationConnectCallback( Widget, XtPointer, XtPointer );
void ComputationIProcCallback( Widget, XtPointer, XtPointer );
void ComputationJProcCallback( Widget, XtPointer, XtPointer );

/********************************************************************\
*                                                                    *
*                            Funktionen                              *
*                            ==========                              *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

void ctrl_init( int *argc, char **argv );
                                /* Initializes the controller              */
void open_X_connection( int *argc, char **argv );      
                                /* opens the connection to the X server    */
void parse_commandline( int *argc, char **argv );  
                                /* extracts the options from commandline   */
void create_windows();          /* opens the windows on screen             */
void register_socket();	        /* Register the socket for asynchronous IO */
				/* at the Xtoolkit			   */
void init_cursor_pixmaps();     /* initializes the cursor pixmaps          */

void ctrl_cleanup();            /* cleans up the controller when the       */
                                /* program is to be finished               */

void cleanup();                 /* terminates the whole program            */

void main_loop();               /* the inner main progam                   */

int everything_saved();         /* Make shure that everything is saved     */

void switch_off_palette();      /* switches off the whole palette */
void switch_on_palette();       /* switches on the palette according to the mode */

XmString fetch_state_string();  /* fetches the label of the current state  */
Pixel fetch_state_color();      /* fetches the color of the current state  */
Pixel fetch_residuum_color();   /* fetches the color of the residuum       */
#endif
