/********************************************************************\
*                                                                    *
*  Projekt:   The Floatator                                          *
*  =============================================                     *
*                                                                    *
*  File:      view.h                                                 *
*                                                                    *
*  Contents:  The definition of the representation specific part     *
*             of the floatator program.                              *
*             This part of the program is concerned with drawing a   *
*             view of the scene by using the primitives of the Xlib. *
*                                                                    *
*  Comment:                                                          *
*                                                                    *
*  History:                                                          *
*                                                                    *
*  Version  Author    Date    Comment                                *
*  ------- -------- -------- -------------------------------         *
*  0.0      Gordon  01.08.95 First steps                             *
*                                                                    *
*                                                                    *
0********1*********2*********3*********4*********5*********6*********7*********8
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

#ifndef VIEW_H
#define VIEW_H

#include "../glib.h"

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>

#include "../problem/model.h"

/********************************************************************\
*                                                                    *
*                         Enum-Konstanten                            *
*                         ===============                            *
*                                                                    *
*  Comment: In alphabetischer Reihenfolge                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von DisplayMode                      *
*                                                                    *
\********************************************************************/

/* Display-Modes fuer die Visualisierung 
 * Legt fest, was mit der Hintergrundfarbe angezeigt wird. */
enum DisplayMode
{ 
    DM_U,	/* Darstellung der horizonalen Geschwindigkeit des Fluids */
    DM_V,	/* Darstellung der vertikalen Geschwindigkeit des Fluids */
    DM_VABS,	/* Darstellung des Geschwindigkeitsbetrages */
    DM_P,	/* Darstellung des Drucks */
    DM_PSI,	/* Darstellung der Stromfunktion */
    DM_ZETA,	/* Darstellung der Wirbelfunktion */
    DM_FLUID	/* Darstellung des Fluid-Typs */
};

/********************************************************************\
*                                                                    *
*                         Datenstrukturen                            *
*                         ===============                            *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  11.07.95  Gordon  Definition von ViewParms                        *
*                                                                    *
\********************************************************************/

/* Voreinstellungen, die aus der Resourcen-Datei gelesen werden. */
/* Farben der einzelnen Komponenten */
struct ViewColors 
{
    Pixel obstacle, empty;
    Pixel fluid[FLUID_FLAVORS];
    Pixel particle[PARTICLE_FLAVORS];
}; 

struct ParticlePixmaps
{
    Pixmap fluid[FLUID_FLAVORS];
    Pixmap particle[PARTICLE_FLAVORS];
};

/* Parameter, die die Darstellung beeinflussen. */
struct ViewParms
{
    REAL xoffset, xwidth, yoffset, yheight;
    REAL zoom;
	/* Wie gross wird eine Zelle dargestellt */
    enum DisplayMode background;    
	/* Was wird im Hintergrund angezeigt */
    int  arrows;		    
	/* Werden Geschwindigkeits-Pfeile dargestellt */
    REAL arrow_length;		    
	/* Wie lang sind die Geschwindigkeitspfeile */
    int particles;
	/* Werden Partikel angezeigt */
    int grid;
	/* Wird ein Gitter angezeigt */
};

/********************************************************************\
*                                                                    *
*                        globale Variablen                           *
*                        =================                           *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  28.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

extern struct ViewColors color;
extern struct ViewParms view;
extern struct ParticlePixmaps pixmaps;
extern GC gc;

/********************************************************************\
*                                                                    *
*                            Funktionen                              *
*                            ==========                              *
*                                                                    *
*  History:                                                          *
*                                                                    *
*    Date    Author   Action                                         *
*  -------- -------- -------------------------------                 *
*  01.08.95  Gordon  Definition                                      *
*                                                                    *
\********************************************************************/

void view_init();
void view_cleanup();

void update_view();

#endif
